# -*- coding: utf-8 -*-
""" Units conversion functions for Eddy Covariance data processing.
This module provides functions to convert gas concentrations from parts per million (ppm) to micromoles per cubic meter (umol/m3),
and to calculate dry air molar concentration based on temperature, pressure, and water vapor content.
"""

# built-in modules
import re
import os
import warnings
import logging
from functools import reduce

# 3rd party modules
import numpy as np
import pandas as pd

# project modules

logger = logging.getLogger('ep._core.units')


def ppm_to_umol_m3(Xgas, Ta, Pa, Xh2o=None):
    """Convert gas concentration from ppm to umol/m3.
    Parameters:
        Xgas (float or pd.Series): Gas concentration in ppm.
        Ta (float or pd.Series): Ambient temperature in Celsius.
        Pa (float or pd.Series): Ambient pressure in KPa.
        Xh2o (float or pd.Series, optional): Water vapor mole fraction in mol/mol. Default is None.
    Returns:
        float or pd.Series: Gas concentration in umol/m3.
    """
    # PV=nRT
    # Pa * m3 = mol * J K-1 mol-1 * K

    if np.quantile(Ta, 0.1) < 100 and np.quantile(Ta, 0.9) < 100:
        Ta = Ta + 274.15
    if np.quantile(Pa, 0.1) < 10 and np.quantile(Pa, 0.9) < 10:
        Pa = Pa * 10**3

    # optical_cell_volume (LICOR 7200: 16 cm3)
    V = 16 / 10**6

    # ℜ = 8.314 J mol-1K-1, the universal gas constant.
    R = 8.374

    # the number of moles inside optical cell (mol)
    n = (Pa * V) / (R * Ta)

    # the number of moles inside optical cell (mol/m3)
    n = (Pa) / (R * Ta)

    if Xh2o is not None:
        # the number of moles of H2O (mol)
        h2o_mols = Xh2o * n
        
        assert h2o_mols < n, 'Error! More mols of water than of air. Check units, water mole fraction should be in mol/mol.'

        # the number of moles inside optical cell wo H2O (mol)
        n_dry = n - h2o_mols
    else:
        n_dry = n

    # the gas concentration (mol.m-3)
    # ppm * n / V 
    return Xgas * n_dry / V

def dry_air_molar_conc(Ta, Pa, Xh2o=None):
    """Calculate the dry air molar concentration based on temperature, pressure, and water vapor content.
    Parameters:
        Ta (float or pd.Series): Ambient temperature in Celsius.
        Pa (float or pd.Series): Ambient pressure in KPa.
        Xh2o (float or pd.Series, optional): Water vapor mole fraction in mol/mol. Default is None.
    Returns:
        float or pd.Series: Dry air molar concentration in mol/m3.
    """
    # Correct temperature (K to °C)
    if np.nanquantile(Ta, 0.1) < 100 and np.nanquantile(Ta, 0.9) < 100:
        Ta = Ta + 274.15
    
    # Correct pressure (KPa to Pa)
    if np.nanquantile(Pa, 0.1) < 10 and np.nanquantile(Pa, 0.9) < 10:
        Pa = Pa * 10**3

    # ℜ = 8.314 J mol-1K-1, the universal gas constant.
    R = 8.374
    
    # Ambient air molar volume
    Va = R * Ta / Pa

    # Molecular weight of water vapour (Mh2o, kgmol-1)
    Mh2o = 0.01802

    # Ambient water vapour mass density (ph2o, kg m-3)
    ph2o = Xh2o * Mh2o / Va

    # water vapor gas constant (Rh2o, JKg-1K-1)
    Rh2o = R / Mh2o

    # Water vapour partial pressure (e, Pa)
    e = ph2o * Rh2o * Ta

    # Dry partial pressure (Pd, Pa)
    Pd = Pa - e
    
    # Dry air molar volume (Vd, m3mol-1)
    Vd = Va * Pa / Pd
    
    return 1 / Vd

def converter(self, Xgas={'co2': 'dryppm'}, Ta='t_cell', Pa='press_cell', Xh2o='h2o', Xh2o_unit=10**-3, drop=True):
    """Convert gas concentrations from ppm to umol/m3.
    Parameters:
        self (pd.DataFrame): DataFrame containing the gas concentrations and other parameters.
        Xgas (dict): Dictionary mapping gas names to their concentrations in ppm or a tuple of (gas_name, concentration).
        Ta (str): Column name for ambient temperature.
        Pa (str): Column name for ambient pressure.
        Xh2o (str): Column name for water vapor mole fraction.
        Xh2o_unit (float): Conversion factor for water vapor mole fraction.
        drop (bool): Whether to drop the original gas concentration columns after conversion. Default is True.
    Returns:
        pd.DataFrame: DataFrame with converted gas concentrations.
    """
    for c, u in Xgas.items():
        if isinstance(u, (list, tuple)):
            #assert(len(u)==2), 'u is too long'
            v, u = u
        else:
            v = c
        if v == c and drop == False:
            self[c+'_bfr_cvt'] = self[c]
            
        self[v] = self[c] * dry_air_molar_conc(Ta=self[Ta], Pa=self[Pa], Xh2o=self[Xh2o]*Xh2o_unit)

    return self


import xarray as xr
from . import conventional_covariance
from .._core import constants

available_methods = {
    'ec':
    type('var_', (object,), {'run': conventional_covariance.covariance,
         'name': 'Conventional covariance.'}),
    'wv':
    type('var_', (object,), {'run': conventional_covariance.covariance,
         'name': 'Wavelet-based covariance (Coimbra et al., 2025)'}),
}

def calculate_L0(data, **kwargs):
    data['FC'] = data['cov_w_co2'] * data['air_molar_volume']**-1
    data['h2o_flux'] = data['cov_w_h2o'] * data['air_molar_volume']**-1
    data['LE'] = data['h2o_flux'] * constants.Mv * constants.Lv
    data['E0'] = data['h2o_flux'] * constants.Mv
    data['H'] = data['cov_w_t_sonic'] * data['rho_m'] * data['cpd']

    data['Tp'] = data['air_temperature'] * (constants.P0a / data['Pa'])**0.286
    data['MO_lenght'] = (data['Tp'] * data['ustar']**3) / \
        (constants.k * constants.g * data['H'] / (data['rho_m'] * data['cpd']))

    return data


def apply_process(data, method=None, **kwargs):
    method = data.attrs.get('process_method', method)
    meth_run = available_methods.get(method, 'ec')

    data = data.mean(dim='time', keep_attrs=True)

    data = calculate_L0(data)

    # for
    # if data[f'cov_{x1}_co2'].data.to_root_units().units == "meter/second":
    #     data[f'{x2}_flux'] = data[f'cov_{x1}_co2'] * \
    #         data['air_molar_volume']**-1

    # if meth_run:
    #     x1 = 'w'
    #     x2 = 'co2'

    #     # needs:
    #     #   w = 'w'
    #     #   x = 'co2', 'h2o', ...
    #     #   corr = data['air_molar_volume']**-1
    #     {'FC': {'base': f'cov_{x1}_co2',
    #             'correction': data['air_molar_volume']**-1}}

    #     # data = meth_run.run(data, x1, x2, **kwargs)

    #     data = data.assign(
    #         **{f'{x2}_flux': data[f'cov_{x1}_co2'] * data['air_molar_volume']**-1,
    #            f'LE': data[f'cov_{x1}_h2o'] * data['air_molar_volume']**-1,
    #            f'H': data[f'cov_{x1}_t_sonic'] * data['air_molar_volume']**-1, }
    #     )
    return data




def fix_time_lag(fix, move, *args, tlag=0, **kwargs):
    return type(
        'var_', (object,),
        {"x": move.shift(time=tlag),# or move.attrs.get('nom_timelag', 0)),
         "tlag": tlag,
         "meta": {}}
        )



"""Despiking functions based on Mauder et al. 2013
"""

# built-in modules
import re
import os
import warnings
import logging
from functools import reduce

# 3rd party modules
import numpy as np
from itertools import islice
from numpy import roll

# project modules
# from ..despiking import window


logger = logging.getLogger('ep.corrections.despiking.mauder_et_al_2013')

def despike(self, cols=['u', 'v', 'w', 'co2', 'h2o'], script='Py', fup2=.1, output_index=None, **kwargs):
    
    plausibility_range = {'u': 3.5, 'v': 3.5, 'w': 5, 'co2': 3.5, 'h2o': 3.5}

    if not callable(script):
        if script=='RFlux':
            script = tcom.LazyCallable(os.path.join(
                cfp.parent, "RFlux-scripts/despiking.R"), "despiking").__get__().fc
            output_index = 0
            kwargs_ = {'mfreq': 20, 'variant': "v3", 'wsignal': 7, 'wscale': 20*60*30/6, 'zth': plausibility_range[c]}
            kwargs_.update(kwargs)
        
        elif script == 'OCE':
            script = tcom.LazyCallable(os.path.join(
                cfp.parent, "corrections/oce_despike.R"), "despike").__get__().fc
            kwargs_ = {'n': plausibility_range[c]}
            kwargs_.update(kwargs)
        
        else:
            script = mauder2013
            kwargs.pop("fill", None)

    for c in cols:
        ogap = np.isnan(self[c])

        # take off absurdity numbers
        p1 = np.nanquantile(self[c], 0.01)
        p99 = np.nanquantile(self[c], 0.99)
        absurdity_bounds_min = 10**(-3 if p1 > 0 else 3) * p1
        absurdity_bounds_max = 10**(3 if p99 > 0 else -3) * p99

        #self.loc[self[c] < absurdity_bounds_min, c] = np.nan
        #self.loc[self[c] > absurdity_bounds_max, c] = np.nan
        print(p1, p99)
        print("np.nanmedian(x)", np.nanmedian(self[c]), np.median(self[c]))
        self[c].mask(self[c] < absurdity_bounds_min, np.nan, inplace=True)
        self[c].mask(self[c] > absurdity_bounds_max, np.nan, inplace=True)
        
        print("np.nanmedian(x)", np.nanmedian(self[c]), np.median(self[c]))

        if output_index:
            self.loc[:, c] = script(self[c], **kwargs)[output_index]
        else:
            self.loc[:, c] = script(self[c], **kwargs)
        
        ngap = np.isnan(self[c])
        N = len(self)
        print(sum(ngap), N, sum(ogap))
        #signan = np.isnan(np.array(self[c]))
        self.loc[:, c] = np.interp(np.linspace(0, 1, N),
                            np.linspace(0, 1, N)[ngap == False],
                            self[c][ngap == False])
        self[c].mask(ogap, np.nan, inplace=True)
        
        #if (np.sum(np.isnan(self[c])) / len(self)) <= fup2:
        #    self.loc[:, c] = self.set_index('TIMESTAMP').loc[:, c].ffill().loc[:, c]

    return self


def mauder2013_(x, n=7, **kwargs):
    for x_i in window(x):
        despiked = x.roll(mauder2013, 7)
    return despiked


def window(seq, n=3):
    """Returns a sliding window (of width n) over data from the iterable
        s = [s(i), s(i+1), ..., s(i+n-1)]
    """
    it = iter(roll(seq, int(n/2)))
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result


def mauder2013(x, q=7, n=7):
    for i, x_ in enumerate(window(
        np.array(x),
        n=n
    )):
        x_ = np.array(x_)
        x_med = float(np.median(x_))
        mad = float(np.median(np.abs(x_ - x_med)))
        bounds = (float(x_med - (q * mad) / 0.6745), 
                float(x_med + (q * mad) / 0.6745))
        # print("median", x_med, "mad", mad, "bounds", bounds)
        if x_[int(n/2)] < min(bounds):
            x[i] = np.nan
        if x_[int(n/2)] > max(bounds):
            x[i] = np.nan

    #if fill is not None:
    #    x = fill(pd.Series(x) if fill in (pd.Series.ffill, pd.Series.interpolate) else x)
    return type('var_', (object,), 
                {"x": x, 
                #  "median": x_med, 
                #  "mad": mad, 
                #  "bounds": bounds,
                 "meta": {}})



import xarray as xr
# from .._core.commons import compute_equation


def covariance(data, x1="w", x2="co2"):
    ds = data.copy()

    # ds = compute_equation(
    #     ds, f'cov_{x1}_{x2}', x1, x2, formula=lambda a, b: xr.cov(a, b), unit_formula=lambda a, b: a * b)
    # content.
    # name_cov_var = f'cov_{x1}_{x2}'
    ds = ds.assign(
        **{f'cov_{x1}_{x2}': xr.cov(data[x1], data[x2], dim='time')}
    )

    # data[name_cov_var].attrs['unit_in'] = ""
    # if 'unit_in' in data[x1].attrs and 'unit_in' in data[x2].attrs:
    #     data[name_cov_var].attrs['unit_in'] = (
    #         ureg(data[x1].attrs['unit_in']) *
    #         ureg(data[x2].attrs['unit_in'])
    #     ).to_root_units()
    return ds



import numpy as np
from .units import convert_unit, ureg
from . import constants

# Constants
Ru = ureg('8.314 J/mol/K')          # universal gas constant J/[mol K]
M_d = ureg('0.02897 kg/mol')       # kg/mol (dry air)
M_h2o = ureg('0.01802 kg/mol')     # kg/mol (water vapor)
R_h2o = Ru / M_h2o     # J/(kg·K)
R_d = Ru / M_d        # J/(kg·K)
e_base = np.e       # base of natural log


def add_micrometeorological_variables_to_data(data):
    # # Example conditions
    # Ta = 298.15   # ambient temperature in K
    # Pa = 101325   # ambient pressure in Pa
    # chi = 0.01    # mol fraction H₂O
    # Tsonic = 299  # sonic temperature if used

    # Ta = data.t_ga.copy()
    # Ta.data = (data.t_ga.values + 273.15) * ureg('kelvin')  # Ambient temperature in Celsius
    Ta = data.ta
    Pa = data.press_cell # * 1e3
    chi = data.h2o * 1e-3
    Tsonic = data.t_sonic

    data = data.assign(Pa=Pa,)

    # T = data.t_sonic - 273.15  # Convert to Celsius if needed
    # if (Stats % Mean(te) > 220d0 . and . Stats % Mean(te) < 340d0) Stats % T = Stats % Mean(te)
    # if (biomet % val(bTa) > 220d0 . and . biomet % val(bTa) < 340d0) Stats % T = biomet % val(bTa)
    # Stats % Pr = Metadata % bar_press
    # if (Stats % Mean(pe) > 40000 . and . Stats % Mean(pe) < 110000) Stats % Pr = Stats % Mean(pe)
    # if (biomet % val(bPa) > 40000 . and . biomet % val(bPa) < 110000) Stats % Pr = biomet % val(bPa)

    # Ambient air molar volume[m+3 mol-1] and air mass density[kg m-3]
    # if (Stats % Pr > 0d0 . and . Stats % T /= error) then
    Va = constants.R * Tsonic / Pa
    data = data.assign(Va=Va,)
    
    data = data.assign(
        Ma=molecular_weight_wet_air(chi),)
    data = data.assign(
        rho_h2ov=rho_h2o(chi, Pa, Va),)
    data = data.assign(
        e=water_vapor_partial_pressure(data.rho_h2ov, data.t_sonic),)

    data = data.assign(
        air_temperature=air_temperature_derived_from_sonic_temperature(data.t_sonic, data.e, data.Pa),)

    data = data.assign(
        es=saturation_vapor_pressure(data.air_temperature),)

    data = data.assign(
        RH=relative_humidity(data.e, data.es),)
    data = data.assign(
        VPD=vapor_pressure_deficit(data.e, data.es),)
    data = data.assign(
        Td=dew_point_temperature(data.e),)  # e in kPa
    data = data.assign(
        Pd=dry_air_partial_pressure(data.Pa, data.e),)
    data = data.assign(
        air_molar_volume=dry_air_molar_volume(data.Pd, data.air_temperature),)
    data = data.assign(
        sealevel_moist_air_molar_volume=sealevel_moist_air_molar_volume(
            data.air_temperature, data.e),)
    data = data.assign(
        rho_d=dry_air_mass_density(data.Pd, Ta),)
    data = data.assign(
        rho_m=moist_air_density(data.rho_d, data.rho_h2ov),)
    data = data.assign(
        cpd=cp_d(Ta),)
    data = data.assign(
        cph2o=cp_h2o(data.RH, Ta),)
    data = data.assign(
        q=specific_humidity(data.e, Pa),)
    data = data.assign(
        Ta_refined=refine_Ta_sonic(data.q, Ta),)
    data = data.assign(
        cp_m=moist_air_cp(data.q, data.cpd, data.cph2o),)
    data = data.assign(
        lambda_v=latent_heat_vaporization(Ta),)
    data = data.assign(
        sigma=density_ratio_sigma(data.rho_h2ov, data.rho_d),)

    return data


def air_temperature_derived_from_sonic_temperature(Ts, e, Pa):
    return Ts * (1+0.32*e/Pa)**-1   # - 273.15


def molecular_weight_wet_air(chi_h2o):
    return constants.Mv * chi_h2o + constants.Md * (1 - chi_h2o)


def rho_h2o(chi_h2o, P, Va):
    # Using ideal gas: ρ = (P * χ) / (R_specific * T)
    # return chi_h2o * P / Va
    return chi_h2o / Va * constants.Mv


def water_vapor_partial_pressure(rho_h2o, Ta):
    # Using ideal gas law: e = ρ * R_specific * T
    # where R_specific for water vapor is R_h2o = Ru / M_h2
    # Ta in Kelvin
    # The same as chi_h2o / Ru * Pa * constants.Mv * R_h2o
    return rho_h2o * constants.Rv * Ta


def saturation_vapor_pressure(Ta):
    # # Campbell & Norman (1998): es = 0.6108 * exp(17.27 * (T - 273.15)/(T - 35.85))
    # Tc = Ta - 273.15
    # return 610.78 * np.exp((17.269 * Tc) / (Tc + 237.3))
    es = Ta.copy()
    Tc = Ta.data.to('delta_degC').magnitude
    es.data = Tc**-8.2 * np.exp(77.345+0.0057*Tc-7235*Tc**-1)
    es.data = es * ureg('Pa')
    return es


def relative_humidity(e, es):
    return 100 * e / es


def vapor_pressure_deficit(e, es):
    return e - es


def dew_point_temperature(e):
    # # Campbell & Norman: Tdew = (243.5 * ln(e/0.6108)) / (17.27 - ln(e/0.6108)) + 273.15
    # return (243.5 * np.log(e_kPa / 0.6108) /
    #         (17.27 - np.log(e_kPa / 0.6108))) + 273.15
    Td = e.copy()
    e_kPa = convert_unit(e, 'kPa').values
    Td.data = 240.97 * np.log(e_kPa / 0.611) / (17.502 - np.log(e_kPa / 0.611))
    Td.data = Td * ureg('kelvin')
    return Td


def dry_air_partial_pressure(Pa, e):
    return Pa - e


def dry_air_molar_volume(Pd, Ta):
    # Ru is the universal gas constant in J/(mol·K)
    # Pd is the dry air partial pressure
    # Ta is the ambient temperature in Celsius
    # Ambient%Vd = (Stats%Pr * Ambient%Va) / Ambient%p_d
    # Va = Ru * Ta / Pa

    # Ta + 273.15
    Vd = constants.R * Ta / Pd
    return Vd


def sealevel_moist_air_molar_volume(Ta, e):
    Vsea = constants.R * Ta / (ureg("99767.5 pascal") - e)
    return Vsea

def dry_air_mass_density(Pd, Ta):
    return Pd / (constants.Rd * Ta)


def moist_air_density(rho_d, rho_h2o):
    return rho_d + rho_h2o


def cp_d(Tk):
    """
    Calculate the specific heat capacity of dry air at constant pressure (cp_d).

    Uses a temperature-dependent empirical formula to compute the specific heat
    capacity of dry air as a function of temperature.
    """
    cpd = Tk.copy()
    # Ta is in celsius
    Ta = Tk.data.to('delta_degC') / ureg('delta_degC')
    cpd.data = (constants.Cpd.to("J/kg/K") +
                ((Ta + 23.12)**2 / 3364) * ureg("J/kg/K"))
    return cpd


def cp_h2o(RH, Tk):
    cpv = Tk.copy()
    # Ta is in celsius
    Ta = Tk.data.to('delta_degC').magnitude
    cpv.data = (1859 + 0.13 * RH + (0.193 + 5.6 * 1e-3 * RH)
                             * Ta + (1e-3 + 5 * 1e-5 * RH) * Ta**2) * ureg("J/kg/K")
    return cpv


def specific_humidity(rho_h2o, rho_a):
    return rho_h2o / rho_a


def refine_Ta_sonic(q, Ta):
    return Ta / (1 + 0.51 * q)


def moist_air_cp(q, cp_d, cp_h2o):
    return (1 - q) * cp_d + q * cp_h2o


def latent_heat_vaporization(Tk):
    # Ta is in celsius
    lambdav = Tk.copy()
    Ta = Tk.data.to('delta_degC')
    lambdav.data = 1e3 * (ureg('3147.5 J/g') -
                          ureg('2.37 J/g/delta_degC') * Ta)
    return lambdav


def density_ratio_sigma(rho_h2o, rho_d):
    return rho_h2o / rho_d


def wind_direction(u, v, offset=0):
    """
    Calculate wind direction from u and v wind components.
    
    Parameters:
        u (float or np.ndarray): zonal wind component (positive eastward)
        v (float or np.ndarray): meridional wind component (positive northward)

    Returns:
        float or np.ndarray: wind direction in degrees, where 0° = North, 90° = East, etc.
    """
    direction = (180 - np.degrees(np.arctan2(v, u))) % 360
    direction = (direction + offset) % 360
    return direction



import pint
from pint.errors import UndefinedUnitError
import pint_xarray

ureg = pint.UnitRegistry(force_ndarray_like=True)
# Define ppm as a dimensionless unit if not already defined
ureg.define('ppm = 1e-6 = parts_per_million')
ureg.define('ppt = 1e-3 = parts_per_thousand')
ureg.define('µmol = 1e-6 mole = micromol')
ureg.define('celsius = delta_degC = celsius')

# Register your registry globally for pint-pandas
pint_xarray.setup_registry(ureg)
pint.set_application_registry(ureg)

# Define aliases (case-insensitive mapping)
UNIT_ALIASES = {
    'ppt': '1e-3',
    'ppm': '1e-6',
    'ppb': '1e-9',
    'kpa': 'kilopascal',
    'μmol': 'micromole',
    'umol': 'micromole',
    'ug': 'microgram',
    'lit/m': 'L/min',
    'celsius': 'delta_degC',
    'kelvin': 'Kelvin',
}


def resolve_unit(unit_str: str):
    """
    Attempts to resolve a unit string using Pint.
    Falls back to alias lookup if the unit is not found.
    """
    unit_str = unit_str.replace('_', '/')

    if not unit_str:
        return ureg('dimensionless')  # Treat empty as dimensionless

    try:
        return ureg(unit_str)
    except UndefinedUnitError:
        # Try alias map (case-insensitive)
        unit_str_lower = unit_str.lower()
        alias = UNIT_ALIASES.get(unit_str_lower)
        if alias:
            return ureg(alias)
        else:
            raise UndefinedUnitError(
                f"Unit '{unit_str}' not found and no alias matched.")


def convert_unit(da, to_units):
    """
    Convert an xarray.DataArray containing Pint Quantities in .values to new units.

    Parameters:
        da: xarray.DataArray with pint.Quantity values
        to_units: string or Pint Unit to convert to, e.g. 'degC', 'meter'

    Returns:
        xarray.DataArray with converted magnitude and updated units attribute.
    """
    q = da.data  # should be a Pint Quantity
    converted_q = q.to(to_units)

    # Rebuild DataArray with converted magnitude and original coords/dims
    da_converted = da.copy()
    da_converted.data = converted_q  # .magnitude
    da_converted.attrs = da.attrs.copy()
    da_converted.attrs['unit_in'] = str(converted_q.units)
    return da_converted


def convert_to_prefered_units(ds, units=["μmol/m²/s", "m/s", "W/m²", "g/m²/s"]):
    for var in ds.data_vars:
        for target_unit in units:
            try:
                # Try to convert
                ds[var].data = ds[var].data.to(target_unit)
                break  # Stop at the first successful conversion
            except Exception as e:
                continue  # Try next unit
    return ds

def DEPRECATED_compute_equation(data, y: str, *args: str, formula, unit_formula=None):
    """
    Compute a new variable `y` using the given formula and attach unit metadata.

    Parameters:
        data: xarray.Dataset or pandas.DataFrame-like (must support .assign and .attrs)
        y (str): Name of the new variable to be created
        *args (str): Names of existing variables used in the formula
        formula (function): A function that takes data[args[0]], data[args[1]], ..., and returns a computed value

    Returns:
        data with new variable y and computed unit metadata
    """
    if not unit_formula:
        unit_formula = formula

    # Evaluate the formula using the columns passed in
    variables = [data[var] for var in args]
    result = formula(*variables)

    # Assign result to new variable y
    data = data.assign(**{y: result})

    # Attempt to compute units
    data[y].attrs = getattr(data[y], 'attrs', {})
    data[y].attrs['unit_in'] = ""

    try:
        # Get unit strings from input variables
        units = [data[arg].attrs['unit_in'] for arg in args]

        # Build Pint quantities for each variable
        pint_quantities = [resolve_unit(u) for u in units]
        # pint_quantities = [ureg(unit if unit else 'dimensionless')
        #                    for unit in units]

        # Simulate formula on Pint quantities to get output unit
        unit_result = unit_formula(*pint_quantities).to_base_units()

        # Store result unit
        data[y].attrs['unit_in'] = str(unit_result)

    except Exception as e:
        data[y].attrs['unit_in'] = f"Error computing units: {e}"

    return data



import pandas as pd


def import_ascii(file_handle='your_binary_file.bin', 
                 head_nlines=1,
                 header_col=0,
                 **kwargs):
    """
    Reads in a generic ASCII file.
    """
    skiprows = kwargs.pop(
        'skiprows', 
        list(set([i for i in range(0, head_nlines)]) - set([header_col]))
    )

    df = pd.read_csv(file_handle, header=0,
                     skiprows=skiprows, **kwargs)
    return df



import tempfile
import xarray as xr
import numpy as np
# pip install cdsapi

def initialize():
    import cdsapi
    c = cdsapi.Client()
    return c


def retrieve_era5(data):
    c = initialize()

    years = list(np.unique(data.date.dt.year).astype(str))
    months = list(np.unique(data.date.dt.month).astype(str))
    days = list(np.unique(data.date.dt.day).astype(str))
    hours = list(np.unique(data.date.dt.hour).astype(str))

    with tempfile.NamedTemporaryFile(suffix='.nc', delete=True) as tmpfile:
        # tmpfile.close()
        # era5_file = tmpfile.name
        c.retrieve(
            'reanalysis-era5-single-levels',
            {
                'product_type': 'reanalysis',
                'variable': 'boundary_layer_height',
                'year': years,
                'month': months,
                'day': days,
                'time': [f'{str(i).zfill(2)}:00' for i in hours],
                'format': 'netcdf',
                # North, West, South, East
                'area': [data.attrs.get('latitude', 0),
                         data.attrs.get('longitude', 0),
                         data.attrs.get('latitude', 0),
                         data.attrs.get('longitude', 0)],
            },
            tmpfile.name
        )

        ds = xr.open_dataset(tmpfile.name)

        ds = ds.rename_dims(valid_time="date").mean(['latitude', 'longitude'])
        ds = xr.DataArray(
            ds['blh'], {'date': ds['valid_time'].rename('date')})

        data = data.assign(blh_era5=ds)

        return data



import xarray as xr
from . import theoretical, fixed, maximisation

available_corrections = {
    'constant':
    type('var_', (object,), {'run': fixed.fix_time_lag,
                             'name': 'constant time lag'}),
    'maxcov&default':
    type('var_', (object,), {'run': maximisation.time_lag_w_default,
                             'name': 'maximization covariance with default'}),
    'maxcov':
    type('var_', (object,), {'run': maximisation.time_lag,
         'name': 'maximization covariance'}),
    # 'tlag_opt':
    # type('var_', (object,), {'run': mauder_et_al_2013.mauder2013,
    #      'name': 'MAD (Mauder, et al., 2013)'}),
}


def apply_time_lag(data, x1='w', x2='co2', **kwargs):
    assert all([v in data for v in [x1, x2]]
               ), 'Not all variables for time lag optimization are in data.'

    correction = available_corrections.get(
        data.attrs.get('Corrections', {}).get('time_lag', {}).get('method', None), None)
    # if correction is a function, then type('var_', (object,), {'run': correction, 'name': ?})

    if correction:
        data = data.assign(**{f'{x2}_bfr_timelag': data[x2]})

        # Translate time lag in sec. to data points
        if 'tlag' not in kwargs:
            kwargs['tlag'] = - int(float(data[x2].attrs.get(
                'nom_timelag', 0)) * float(data.attrs.get('acquisition_frequency', 0)))
            kwargs['tlag_min'] = - int(float(data[x2].attrs.get(
                'min_timelag', 0)) * float(data.attrs.get('acquisition_frequency', 0)))
            kwargs['tlag_max'] = - int(float(data[x2].attrs.get(
                'max_timelag', 0)) * float(data.attrs.get('acquisition_frequency', 0)))

        dr = correction.run(data[x1], data[x2], **kwargs)

        data[x2].data = dr.x

        data[x2] = data[x2].assign_attrs(time_lag_method=correction.name)
        for k, v in vars(dr).items():
            if k.startswith('__') or (k in ['x', 'meta']):
                continue
            else:
                try:
                    if k in ['tlag']:
                        try:
                            v = abs(v * \
                                float(data.attrs.get('acquisition_frequency', 0)) ** -1)
                        except:
                            pass
                    var_name = f'time_lag_params_{k}_{x2}'
                    
                    # logger.debug(f'Assigning {var_name} with value {v} to data.')
                    data = data.assign(**{var_name: v})
                    data[var_name] = data[var_name].assign_attrs(
                        time_lag_method=correction.name)
                except:
                    pass

        if 'meta' in vars(dr).keys() and dr.meta:
            data[x2] = data[x2].assign_attrs(time_lag_meta=dr.meta)

        data.attrs['corrections_applied'] += f'time lag optimization w/ {correction.name}; '
    return data



# the following line needed for unicode character in convert_anglestring
# -*- coding: latin-1 -*-
# standard modules
import copy
import datetime
import logging
import numbers
import os
import platform
import sys
import time
# third party modules
import dateutil
import numpy
import pytz
import xlrd
# Local modules

logger = logging.basicConfig()

def get_base_path():
    """
    Purpose:
     Return the base path dependng on whether we are running as a script
     or a Pyinstaller application.
    Author: https://stackoverflow.com/questions/7674790/bundling-data-files-with-pyinstaller-onefile
    """
    # check if we running as a PyInstaller application
    if getattr(sys, 'frozen', False):
        # running as a PyInstaller application
        base_path = sys._MEIPASS
    else:
        # running as a script
        base_path = os.path.abspath(".")
    return base_path


def file_exists(filename, mode="quiet"):
    """
    Purpose:
     Return True or False if file exists
    Author: OzFlux
    """
    if not os.path.exists(filename):
        if mode != "quiet":
            logger.error(" File " + filename + " not found")
        return False
    else:
        return True


def update_nested_dict(d, u):
    """
    Recursively updates a nested dictionary `d` with values from another dictionary `u`.
    If a key in `u` maps to a dictionary and the corresponding key in `d` also maps to a dictionary,
    the function updates the nested dictionary in `d`. Otherwise, it overwrites the value in `d`.

    Args:
        d (dict): The dictionary to update.
        u (dict): The dictionary containing updates.

    Returns:
        dict: The updated dictionary.
    """
    # Iterate over each key-value pair in the update dictionary `u`
    for k, v in u.items():
        # Check if the current value is a dictionary
        if isinstance(v, dict):
            # If the corresponding value in `d` is also a dictionary, recursively update it
            # Use `d.get(k, {})` to handle cases where the key `k` is not already in `d`
            d[k] = update_nested_dict(d.get(k, {}), v)
        else:
            # If the value is not a dictionary, directly update/overwrite the key in `d`
            d[k] = v
    # Return the updated dictionary
    return d


def update_nested_dicts(*ds, fstr=None):
    r = {}
    for d in ds:
        if isinstance(d, str) and fstr:
            try:
                d = fstr(d)
            except Exception as e:
                continue
        r = update_nested_dict(r, d)
    return r



import pandas as pd
import xarray as xr
import os


def export_as_ONEFlux(ds: xr.Dataset, output_file='fluxnet_output.csv'):
    header = []
    header += [f"site,{ds.attrs['site_name']}"]
    # header += [f"year,{ds.attrs['latitude']}"]
    header += [f"lat,{ds.attrs['latitude']}"]
    header += [f"lon,{ds.attrs['longitude']}"]
    header += [f"timezone,?"]
    header += [f"htower,?"]
    header += [f"htower,?"]
    header += [f"timeres,halfhourly"]
    header += [f"sc_negl,1"]
    header += [f"notes,{pd.Timestamp.now().strftime('%Y%m%d%H%M')} qc visual comparison SY"]
    header += [f"notes,dataset retrieved on {pd.Timestamp.now().strftime('%Y%m%d%H%M%S')}"]

    # 1. Variable mapping
    rename_map = {
        # GASES
        'co2': 'CO2',
        'h2o': 'H2O',
        'ch4': 'CH4',
        'no': 'NO',
        'no2': 'NO2',
        'n2o': 'N2O'
        }
    
    expected_columns = ['CO2','FC','G','H','H2O',
                        'LE','NEE_pi','NETRAD','P','PA','PPFD_IN','RH','SWC_1',
                        'SW_IN','TA','TS_1','USTAR','VPD','WD','WS']

    ds = ds.rename({k: k.upper() for k in ds.data_vars})

    csv_data = pd.DataFrame({
        'TIMESTAMP_START': ds['date'] - ds.date.attrs['delta'],
        'TIMESTAMP_END': ds['date'],
    })

    for col in expected_columns:
        csv_data[col] = ds[col].data.magnitude.ravel() if col in ds else pd.NA

    # Write .csv
    csv_header = "\n".join(header)

    # Guarantee folders exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write to file
    with open(output_file, 'w', newline='') as f:
        # First write the pre-data lines
        f.write(csv_header + '\n')
        csv_data.to_csv(f, index=False)
    return


import logging
import xarray as xr
from . import conventional_covariance, L0_fluxes
from .._core import constants
from .. import version, _core, io, process, corrections, qaqc, external, compatibility
from .._core.units import convert_to_prefered_units
# from concurrent.futures import ProcessPoolExecutor

logger = logging.getLogger()


def preprocess_eddy_covariance_data(data):
    data = _core.micrometeorology.add_micrometeorological_variables_to_data(
        data)

    data = data.assign(
        wind_dir=_core.micrometeorology.wind_direction(
            data.u.mean(), data.v.mean(), float(data.attrs.get('anemometer_north_offset', 0))),)

    logger.debug('Micrometeorology added to data. Starting corrections.')

    data = data.pint.dequantify()
    data = corrections.despiking.apply_despiking(data, ['u', 'v', 'w', 'co2'])
    data = corrections.time_lag.apply_time_lag(data)
    data = corrections.axis_rotation.apply_axis_rotation(data)
    data = data.pint.quantify()

    logger.debug(
        'High frequency corrections done. Starting covariance calculations.')

    data_sta = data[["u", "v", "w", "co2", "h2o"]].std()
    data_sta = data_sta.rename(
        {k: f'std_{k}' for k in list(data_sta.variables.keys())})
    data = xr.merge([data, data_sta])

    for x1, x2 in [('w', 'co2'), ('w', 'h2o'), ('w', 't_sonic'), ('u', 'v'), ('u', 'w')]:
        data = data.assign(
            **{f'cov_{x1}_{x2}': xr.cov(data[x1], data[x2])}
        )

    data = data.assign(ustar=(data.cov_u_w.pipe(abs) ** 0.5))

    logger.debug('Covariances calculated. Calculating quality control.')

    data = data.pint.dequantify()
    data = qaqc.sta(data)
    data = data.pint.quantify()

    logger.debug('All pre processing done.')

    data = convert_to_prefered_units(data)

    # Convert timestamp to datetime and set as index
    # data = data.assign_coords(date=data['timestamp_end'], time=data['time_ns'])
    # data = data.set_index(timestamp=["date", "time"]).unstack("timestamp")
    return data


def process_time_series(setup, datetimes, keep_raw=False):
    raw_data = io.importer.import_data(
        setup, datetimes, io.importer.import_ascii.import_ascii)

    data = io.importer.read_data.format_data(
        raw_data, corrections_applied='', **setup.copy())

    data = data.groupby(['date', 'latitude', 'longitude']
                        ).apply(preprocess_eddy_covariance_data)

    processed_data = data.mean(dim='time', keep_attrs=True)

    if keep_raw:
        # Only keep variables that have the 'time' dimension
        data_with_time = data[[
            var for var in data.data_vars if 'time' in data[var].dims]]

        # Concatenate along the 'date' dimension
        processed_data = xr.concat(
            [processed_data, data_with_time], dim='date')
    # if keep_raw = True:
    #   data = only the data which has 'time' in dims
    #   xr.concat([processed_data, data], dim='date')

    processed_data = process.L0_fluxes.calculate_L0(processed_data)

    processed_data = _core.units.convert_to_prefered_units(processed_data)

    return processed_data


def process_time_series_in_windows(setup, parallel=False, **kwargs):
    time_range = setup['datetimerange']
    file_freq = f"{setup['Files']['fileduration']}Min"
    process_freq = f"3h"

    start_time, end_time = io.importer.read_data.d0_d1_from_time_range(
        time_range)
    time_windows = io.importer.read_data.generate_time_windows(
        start_time, end_time, file_freq, process_freq)

    processed_chunks = []

    # # Parallel execution
    # if parallel:
    #     # Build args list
    #     arg_list = [(setup, w) for w in time_windows]
    #     with ProcessPoolExecutor() as executor:
    #         processed_chunks = list(executor.map(
    #             process_time_series, arg_list))
    # else:
    for window in time_windows:
        processed_data = process_time_series(setup, window, **kwargs)

        processed_chunks.append(processed_data)

    final_dataset = xr.concat(processed_chunks, dim='date')
    return final_dataset



import os
import numpy as np
# from scripts import common as tcom
import pathlib
import warnings
from scipy import optimize
import xarray as xr
# current file path
# cfp = pathlib.Path(__file__).parent.resolve()

def default_lag(length=711, diameter=5.3, pump=15, dt=20):
    # default time lag in number of data points
    return int(np.round((length * (np.pi * (diameter/2)**2) * (10**-6) / pump) * 60 * dt))


def cost_function(fix, move, shift):
    shift = int(np.round(shift))
    return -np.abs(xr.cov(fix, move.shift(time=shift)))


def cost_function_w_bounds(fix, move, shift, shift_min, shift_max):
    shift = int(np.round(shift))
    if (shift_min and shift < shift_min) or (shift_max and shift > shift_max):
        return 9999
    return -np.abs(xr.cov(fix, move.shift(time=shift)))


def time_lag(fix, move, tlag, **kwargs):
    opt = optimize.minimize(
        lambda t: cost_function(fix, move, t), [tlag])
    moved = move.shift(time=int(np.round(opt.x)))

    return type(
        'var_', (object,),
        {"x": moved,
         "tlag": int(opt.x),
         "meta": {}}
    )


def time_lag_w_default(fix, move, tlag, tlag_min=0, tlag_max=100, **kwargs):
    opt = optimize.brute(
        lambda t: cost_function(fix, move, t), [(tlag_min, tlag_max)], Ns=(tlag_max - tlag_min + 1),)
    
    opt_use = int(np.round(opt))
    if (tlag_min and np.abs(opt_use) < np.abs(tlag_min)):
        opt_use = -tlag_min
    elif (tlag_max and np.abs(opt_use) > np.abs(tlag_max)):
        opt_use = -tlag_max

    moved = move.shift(time=opt_use)

    return type(
        'var_', (object,),
        {"x": moved,
         "tlag_opt": float(opt),
         "tlag": opt_use,
         "meta": {}}
    )


def time_lag_w_default_opt(fix, move, tlag, tlag_min=None, tlag_max=None, **kwargs):
    opt = optimize.basinhopping(
        lambda t: cost_function_w_bounds(fix, move, t, tlag_min, tlag_max), [tlag], stepsize=1)
    moved = move.shift(time=int(opt.x))

    return type(
        'var_', (object,),
        {"x": moved,
         "tlag": int(opt.x),
         "meta": {}}
    )




# built-in modules
from configobj import ConfigObj
import re
import os
import warnings
import logging
from functools import reduce

# 3rd party modules
import numpy as np
import pandas as pd

# project modules

logger = logging.getLogger('wvlt.oneflux_compatibility')


def write_files_as_01_qc_visual(config, dst, overwrite=True):
    template = [
        f"site, {config.get('site_name', '')}",
        f"year, 2005",
        f"lat, {config.get('latitude', '')}",
        f"lon, {config.get('longitude', '')}",
        f"timezone, -6",
        f"htower, 200501010030, 4.05",
        f"timeres, halfhourly",
        f"sc_negl, 1"]
    # TIMESTAMP_START, TIMESTAMP_END, CO2, FC, G, H, H2O, LE, NEE_pi, NETRAD, P, PA, PPFD_IN, RH, SWC_1, SW_IN, TA, TS_1, USTAR, VPD, WD, WS
    
    if os.path.exists(dst) and not overwrite:
        logger.info(f"File `{dst}` already exists, to overwrite state `overwrite=True`.")
        return
    
    with open(dst, 'w+') as wf:
        wf.write("\n".join(template))
    return



import pandas as pd
import xarray as xr


def export_as_FLUXNET(ds: xr.Dataset, output_file='fluxnet_output.csv'):
    """
    Export an xarray.Dataset to a FLUXNET-style CSV file, with variable renaming,
    timestamp formatting, and unit verification/conversion.

    Parameters:
    - ds: xarray.Dataset
    - output_file: str
    """

    # 1. Variable mapping
    rename_map = {
        # GASES
        'co2': 'CO2',
        'h2o': 'H2O',
        'ch4': 'CH4',
        'no': 'NO',
        'no2': 'NO2',
        'n2o': 'N2O',
        'o3': 'O3',
        'co2_flux': 'FC',
        'ch4_flux': 'FCH4',
        'no_flux': 'FNO',
        'no2_flux': 'FNO2',
        'n2o_flux': 'FN2O',
        'fo3_flux': 'FO3',
        'sc': 'SC',
        'sch4': 'SCH4',
        'sno': 'SNO',
        'sno2': 'SNO2',
        'sn2o': 'SN2O',
        'so3': 'SO3',

        # FOOTPRINT
        'fetch_max': 'FETCH_MAX',
        'fetch_90': 'FETCH_90',
        'fetch_55': 'FETCH_55',
        'fetch_40': 'FETCH_40',
        'fetch_filter': 'FETCH_FILTER',

        # HEAT
        'g': 'G',
        'h': 'H',
        'le': 'LE',
        'netrad': 'NETRAD',

        # MET
        'air_temperature': 'TA',
        'rh': 'RH',
        'Pa': 'PA',
        'vpd': 'VPD',
        't_sonic': 'TS',
        'swc': 'SWC',
        'p': 'P',
        'p_rain': 'P_RAIN',

        # RADIATION
        'sw_in': 'SW_IN',
        'sw_out': 'SW_OUT',
        'lw_in': 'LW_IN',
        'lw_out': 'LW_OUT',
        'ppfd_in': 'PPFD_IN',
        'ppfd_out': 'PPFD_OUT',
        'apar': 'APAR',

        # PRODUCTS
        'nee': 'NEE',
        'reco': 'RECO',
        'gpp': 'GPP',

        # WIND
        'u': 'WS',
        'wind_dir': 'WD',
        'ustar': 'USTAR',
    }

    # Filter only the variables available in the dataset
    rename_map_filtered = {k: v for k,
                           v in rename_map.items() if k in ds.data_vars}

    # Rename variables to FLUXNET standard names
    ds = ds.rename(rename_map_filtered)

    # 2. Convert dataset to DataFrame
    df = ds.to_dataframe().reset_index()

    # 3. Time formatting
    if 'time' in df.columns:
        df['TIMESTAMP'] = df['time'].dt.strftime(
            '%Y%m%d%H%M%S')  # FLUXNET wants full timestamp
        df = df.drop(columns=['time'])
        df = df[['TIMESTAMP'] + [col for col in df.columns if col != 'TIMESTAMP']]

    # 4. Unit conversion placeholder — implement actual checks per variable if needed
    # Example: if your temperature is in Kelvin, convert to deg C
    if 'TA' in df.columns and df['TA'].max() > 100:  # crude check for Kelvin
        df['TA'] = df['TA'] - 273.15

    # Add other unit conversions here if your data needs it
    # For example, if RH is in [0–1], multiply by 100
    if 'RH' in df.columns and df['RH'].max() <= 1:
        df['RH'] = df['RH'] * 100

    # 5. Export CSV
    df.to_csv(output_file, index=False)
    print(f"FLUXNET-compatible CSV exported: {output_file}")

    return



from . import PyFluxPro, EddyPro, OneFlux
from configobj import ConfigObj


def convert_to_setup(setup, save_to=None):
    setup = ConfigObj(setup)
    setup["level"] = "L0"
    setup.filename = save_to
    if save_to:
        setup.write()
    return setup



import pandas as pd
import xarray as xr
from . import version, _core, io, process, corrections,  qaqc, external, compatibility
# ignore gapfilling, partitioning, footprint,
from ._core import utils
from .process import process_time_series_in_windows, process_time_series, preprocess_eddy_covariance_data
# from . import version, handler, _core, corrections, gapfilling, partitioning, footprint
# from ._core.eddycov import universal_wt as wavelet_transform
# from .handler import run_from_eddypro
# from ._core.pipeline import process, main




version_name = "oneflux_preproc"
__name__ = version_name.lower().strip().replace(' ', '_')
version_number = "V0.0.0.0.1"
__version__ = version_number
full_version = version_number

release = 'dev' not in version_number and '+' not in version_number
short_version = version_number.split("+")[0]



# standard modules
# 3rd party modules
import numpy
# PFP modules
from scripts import constants as c
from scripts import pfp_utils

def absolutehumidityfromrelativehumidity(Ta,RH):
    # convert to masked arrays
    RH, WasND = pfp_utils.SeriestoMA(RH)
    Ta, dummy = pfp_utils.SeriestoMA(Ta)
    # do the job
    vp = RH * VPsat(Ta) / float(100)
    ah = float(1000000) * vp / ((Ta + 273.15) * c.Rv)
    # convert back to ndarray if input is not a masked array
    if WasND: ah, _ = pfp_utils.MAtoSeries(ah)
    return ah

def co2_ppmfrommgCO2pm3(c_mgpm3,T,p):
    """
     Convert CO2 mass density (mgCO2/m^3) to mole fraction (umol/mol)
        Usage:
         CO2_ppm = co2_ppmfrommgCO2pm3(CO2_mgpm3, T, p)
         where
         CO2_mgpm3 (input) - CO2 concentration, mgCO2/m^3
         T (input) - air temperature, degC
         p (input) - air pressure, kPa
        Returns the CO2 mole fraction in umol/mol.
    """
    # convert to masked array if required
    c_mgpm3, WasND = pfp_utils.SeriestoMA(c_mgpm3)
    T, dummy = pfp_utils.SeriestoMA(T)
    T = T + 273.15             # temperature in K
    p, dummy = pfp_utils.SeriestoMA(p)
    p = p * float(1000)        # pressure in Pa
    # do the job
    c_ppm = (c_mgpm3/c.Mco2)*c.R*T/p
    # convert back to ndarray if input is not a masked array
    if WasND: c_ppm, _ = pfp_utils.MAtoSeries(c_ppm)
    return c_ppm

def co2_mgCO2pm3fromppm(c_ppm,T,p):
    """
     Convert CO2 concentration units of umol/mol (ppm) to mgCO2/m3
        Usage:
         CO2_mgpm3 = co2_mgCO2pm3fromppm(CO2_ppm, T, p)
         where
         CO2_ppm (input) - CO2 concentration, umol/mol
         T (input) - air temperature, C
         p (input) - air pressure, kPa
        Returns the CO2 concentration in mg/m^3.
    """
    # convert to masked array if required
    c_ppm, WasND = pfp_utils.SeriestoMA(c_ppm)
    T, dummy = pfp_utils.SeriestoMA(T)
    T = T + 273.15             # temperature in K
    p, dummy = pfp_utils.SeriestoMA(p)
    p = p * float(1000)        # pressure in Pa
    # do the job
    c_mgpm3 = c_ppm*c.Mco2*p/(c.R*T)
    # convert back to ndarray if input is not a masked array
    if WasND: c_mgpm3, _ = pfp_utils.MAtoSeries(c_mgpm3)
    return c_mgpm3

def co2_umolpm3fromppm(c_ppm, T, p):
    """
     Convert CO2 concentration units of umol/mol (ppm) to umol/m^3
        Usage:
         CO2_umolpm3 = co2_umolpm3fromppm(CO2_ppm, T, p)
         where
         CO2_ppm (input) - CO2 concentration, umol/mol
         T (input) - air temperature, C
         p (input) - air pressure, kPa
        Returns the CO2 concentration in umol/m^3.
    """
    # convert to masked array if required
    c_ppm, WasND = pfp_utils.SeriestoMA(c_ppm)
    T, dummy = pfp_utils.SeriestoMA(T)
    T = T + 273.15             # temperature in K
    p, dummy = pfp_utils.SeriestoMA(p)
    p = p * float(1000)        # pressure in Pa
    # do the job
    c_umolpm3 = c_ppm*p/(c.R*T)
    # convert back to ndarray if input is not a masked array
    if WasND: c_umolpm3, _ = pfp_utils.MAtoSeries(c_umolpm3)
    return c_umolpm3

def densitydryair(Ta,ps,vp):
    # Calculate density of dry air from temperature, pressure and vapour pressure
    #  Ta - air temperature, C
    #  ps - pressure, kPa
    #  vp - vapour pressure, kPa
    # Returns
    #  rhod - dry air density, kg/m3
    rhod = 1000*(ps-vp)/((Ta+273.15)*c.Rd)
    return rhod

def densitymoistair(Ta,ps,vp):
    # Calculate density of moist air from temperature, pressure and vapour pressure
    #  Ta - air temperature, C
    #  ps - pressure, kPa
    #  vp - vapour pressure, kPa
    # Returns
    #  rhom - moist air density, kg/m3
    rhow = vp*1000/((Ta+273.15)*c.Rv)
    rhod = (ps-vp)*1000/((Ta+273.15)*c.Rd)
    rhom = rhow + rhod
    return rhom

def densitywatervapour(Ta,vp):
    # Calculate partial density of water vapour from temperature and vapour pressure
    #  Ta - air temperature, C
    #  vp - vapour pressure, kPa
    # Returns
    #  rhow - partial density of water vapour, kg/m3
    rhow = vp*1000/((Ta+273.15)*c.Rv)
    return rhow

def VPsat(T):
    # Saturation vapour pressure.
    #  T is the air temperature, C
    #  VPsat is the saturation vapour pressure in kPa
    return 0.6106 * numpy.exp(17.27 * T / (T + 237.3))

def ET_kgpm2fromkgpm2ps(ET_kgpm2ps, ts):
    """
    Convert ET in units of kg/m^2/s to units of kg/m^2
    Usage:
     ET_kgpm2ps = ET_kgpm2fromkgpm2ps(ET_kgpm2ps, ts)
     where:
      ET_kgpm2ps (input) - ET in units of kg/m^2/s
      ts - time step in minutes
    Returns ET in units of kg/m^2
    """
    # convert to masked array
    ET_kgpm2ps, WasND = pfp_utils.SeriestoMA(ET_kgpm2ps)
    # do the job
    ET_kgpm2 = ET_kgpm2ps*ts*60
    # convert back to ndarray if input is not a masked array
    if WasND: ET_kgpm2, _ = pfp_utils.MAtoSeries(ET_kgpm2)
    return ET_kgpm2

def Fco2_gCpm2fromumolpm2ps(Fc_umolpm2ps, ts):
    """
    Convert Fco2 in units of umol/m^2/s to units of gC/m^2 (C, not CO2)
    Usage:
     Fco2_gCpm2ps = Fco2_gCpm2fromumolpm2ps(Fc_umolpm2ps, ts)
     where:
      Fco2_umolpm2ps (input) - CO2 flux in units of umol/m^2/s
      ts - time step in minutes
    Returns the CO2 flux in units of gC/m^2 (C, not CO2)
    """
    # convert to masked array
    Fc_umolpm2ps, WasND = pfp_utils.SeriestoMA(Fc_umolpm2ps)
    # do the job
    Fc_gCpm2 = Fc_umolpm2ps*12.01*ts*60/1E6
    # convert back to ndarray if input is not a masked array
    if WasND: Fc_gCpm2, _ = pfp_utils.MAtoSeries(Fc_gCpm2)
    return Fc_gCpm2

def Fco2_gCpm2psfromumolpm2ps(Fc_umolpm2ps):
    """
    Convert Fco2 in units of umol/m^2/s to units of g/m^2/s (C, not CO2)
    Usage:
     Fco2_mgpm2ps = Fco2_gCpm2psfromumolpm2ps(Fc_umolpm2ps)
     where:
      Fco2_umolpm2ps (input) - CO2 flux in units of umol/m^2/s
    Returns the CO2 flux in units of g/m^2/s (C, not CO2)
    """
    # convert to masked array
    Fc_umolpm2ps, WasND = pfp_utils.SeriestoMA(Fc_umolpm2ps)
    # do the job
    Fc_gCpm2ps = Fc_umolpm2ps * c.Mc/1E3
    # convert back to ndarray if input is not a masked array
    if WasND: Fc_gCpm2ps, _ = pfp_utils.MAtoSeries(Fc_gCpm2ps)
    return Fc_gCpm2ps

def Fco2_mgCO2pm2psfromumolpm2ps(Fc_umolpm2ps):
    """
    Convert Fc in units of umol/m^2/s to units of mgCO2/m2/s
    Usage:
     Fc_mgCO2pm2ps = Fco2_mgCO2pm2psfromumolpm2ps(Fc_umolpm2ps)
     where:
      Fc_umolpm2ps (input) - CO2 flux in units of umol/m^2/s
    Returns the CO2 flux in units of mgCO2/m2/s
    """
    # convert to masked array
    Fc_umolpm2ps, WasND = pfp_utils.SeriestoMA(Fc_umolpm2ps)
    # do the job
    Fc_mgCO2pm2ps = Fc_umolpm2ps * c.Mco2
    # convert back to ndarray if input is not a masked array
    if WasND: Fc_mgCO2pm2ps, _ = pfp_utils.MAtoSeries(Fc_mgCO2pm2ps)
    return Fc_mgCO2pm2ps

def Fco2_umolpm2psfromgCpm2(Fc_gCpm2, ts=None):
    """
    Convert Fc in units of gC/m^2 to units of umol/m^2/s
    Usage:
     Fc_umolpm2ps = Fco2_umolpm2psfromgCpm2(Fc_gCpm2)
     where:
      Fc_gCpm2 (input) - CO2 flux in units of gC/m^2 per time step
    Returns the CO2 flux in units of umol/m^2/s
    """
    # convert to masked array
    Fc_gCpm2, WasND = pfp_utils.SeriestoMA(Fc_gCpm2)
    # do the job
    Fc_umolpm2ps = Fc_gCpm2*1E6/(12.01*ts*60)
    # convert back to ndarray if input is not a masked array
    if WasND: Fc_umolpm2ps, _ = pfp_utils.MAtoSeries(Fc_umolpm2ps)
    return Fc_umolpm2ps

def Fco2_umolpm2psfrommgCO2pm2ps(Fc_mgpm2ps):
    """
    Convert Fc in units of mg/m^2/s to units of umol/m^2/s
    Usage:
     Fc_umolpm2ps = Fco2_umolpm2psfrommgCO2pm2ps(Fc_mgpm2ps)
     where:
      Fc_mgpm2ps (input) - CO2 flux in units of mg/m^2/s
    Returns the CO2 flux in units of umol/m^2/s
    """
    # convert to masked array
    Fc_mgpm2ps, WasND = pfp_utils.SeriestoMA(Fc_mgpm2ps)
    # do the job
    Fc_umolpm2ps = Fc_mgpm2ps / c.Mco2
    # convert back to ndarray if input is not a masked array
    if WasND: Fc_umolpm2ps, _ = pfp_utils.MAtoSeries(Fc_umolpm2ps)
    return Fc_umolpm2ps

def h2o_gpm3frommmolpmol(h_mmpm,T,p):
    """
     Convert H2O concentration units of mmol/mol to g/m^3.
        Usage:
         H2O_gpm3 = h2o_gpm3frommmolpmol(H2O_mmolpmol, T, p)
         where
         H2O_mmolpmol (input) - H2O concentration, mmol/mol
         T (input) - air temperature, C
         p (input) - air pressure, kPa
        Returns the H2O concentration in g/m^3.
    """
    # convert to masked arrays
    h_mmpm, WasND = pfp_utils.SeriestoMA(h_mmpm)
    T, dummy = pfp_utils.SeriestoMA(T)
    p, dummy = pfp_utils.SeriestoMA(p)
    # do the job
    h_gpm3 = (c.Mv*h_mmpm*p*1000)/(c.R*(T+273.15))
    # convert to ndarray if input is not a masked array
    if WasND: h_gpm3, _ = pfp_utils.MAtoSeries(h_gpm3)
    return h_gpm3

def h2o_mmolpmolfromgpm3(h_gpm3,T,p):
    """
     Convert H2O concentration units of g/m^3 to mmol/mol.
        Usage:
         H2O_mmolpmol = h2o_mmolpmolfromgpm3(H2O_gpm3, T, p)
         where
         H2O_gpm3 (input) - H2O concentration, g/m^3
         T (input) - air temperature, C
         p (input) - air pressure, kPa
        Returns the H2O concentration in mmol/mol.
    """
    # convert to masked arrays
    h_gpm3, WasND = pfp_utils.SeriestoMA(h_gpm3)
    T, dummy = pfp_utils.SeriestoMA(T)
    p, dummy = pfp_utils.SeriestoMA(p)
    # do the job
    h_mmpm = (h_gpm3/c.Mv)*c.R*(T+273.15)/(p*1000)
    # convert to ndarray if input is not a masked array
    if WasND: h_mmpm, _ = pfp_utils.MAtoSeries(h_mmpm)
    return h_mmpm

def h2o_mmolpm3fromgpm3(h_gpm3):
    """
     Convert H2O concentration units of g/m^3 to mmol/m^3.
        Usage:
         H2O_mmolpm3 = h2o_mmolpm3fromgpm3(H2O_gpm3)
         where
         H2O_gpm3 (input) - H2O concentration, g/m^3
        Returns the H2O concentration in mmol/m^3.
    """
    # convert to masked arrays
    h_gpm3, WasND = pfp_utils.SeriestoMA(h_gpm3)
    # do the job
    h_mmolpm3 = h_gpm3/float(c.Mv)
    # convert to ndarray if input is not a masked array
    if WasND: h_mmolpm3, _ = pfp_utils.MAtoSeries(h_mmolpm3)
    return h_mmolpm3

def Lv(Ta):
    # Calculate Lv as a function of temperature, from Stull 1988
    #  Ta - air temperature, C
    # Returns
    #  Lv - latent heat of vapourisation, J/kg
    return 2500800 - (2366.8 * Ta)

def mixingratio(ps,vp):
    # Calculate mixing ratio from vapour pressure and pressure
    #  ps - presure, kPa
    #  vp - vapour pressure, kPa
    # Returns
    #  mr - mixing ratio, kg/kg
    return 0.622*vp/(ps- vp)

def molen(T,AH,p,ustar,heatflux,fluxtype='sensible'):
    # Calculate the Monin-Obukhov length
    ustar = numpy.ma.sqrt(numpy.square(ustar))    # force the sign of ustar to be positive
    vp = vapourpressure(AH,T)       # calculate the vapour pressure
    mr = mixingratio(p,vp)          # calculate the mixing ratio
    Tv = theta(T,p)                 # calculate potential temperature
    Tvp = virtualtheta(Tv,mr)
    if fluxtype=='sensible':
        L = -Tvp*densitydryair(T, p, vp)*c.Cp*(ustar**3)/(c.g*c.k*heatflux)
    elif fluxtype=='kinematic':
        L = -Tvp*(ustar**3)/(c.g*c.k*heatflux)
    else:
        raise Exception(" meteorologicalfunctions.molen: unkown value for fluxtype (="+str(fluxtype)+") encountered")
    return L

def specifichumidityfromrelativehumidity(RH, T, p):
    # Specific humidity (kg/kg) from relative humidity, temperature and pressure
    #  RH is the relative humidity, percent
    #  T is the air temperature, degC
    #  p is the atmospheric pressure, kPa
    SH = (c.Mv / c.Md) * (0.01 * RH * VPsat(T) / p)
    return SH

def SHsat(vpsat, ps):
    return 0.622 * (vpsat / ps)

def relativehumidityfromabsolutehumidity(AH, Ta):
    # Relative humidity from absolute humidity
    #  Ta is the air temperature, degC
    #  AH is the absolute humidity, g/m^3
    #  RH is the relative humidity, percent
    # convert to masked arrays
    AH, WasND = pfp_utils.SeriestoMA(AH)
    Ta, dummy = pfp_utils.SeriestoMA(Ta)
    # do the job
    VP = vapourpressure(AH, Ta)
    RH = float(100)*VP/VPsat(Ta)
    # convert back to ndarray if input is not a masked array
    if WasND: RH, _ = pfp_utils.MAtoSeries(RH)
    return RH

def relativehumidityfromdewpoint(Td,Ta):
    # Relative humidity from dew point temperature
    #  Ta is the air temperature, C
    #  Td is the dew point temperature, C
    #  RH is the relative humidity, %
    # convert to masked arrays
    Td, WasND = pfp_utils.SeriestoMA(Td)
    Ta, dummy = pfp_utils.SeriestoMA(Ta)
    # do the job
    RH = 100*10**(7.591386*(Td/(Td+240.7263)-Ta/(Ta+240.7263)))
    # convert back to ndarray if input is not a masked array
    if WasND: RH, _ = pfp_utils.MAtoSeries(RH)
    return RH

def relativehumidityfromspecifichumidity(SH,Ta,ps):
    # Relative humidity from specific humidity
    #  SH is the specific humidity, kg/kg
    #  Ta is the air temperature, degC
    #  ps is the pressure, kPa
    #  RH is the relative humidity, percent
    # convert to masked arrays
    SH, WasND = pfp_utils.SeriestoMA(SH)
    Ta, dummy = pfp_utils.SeriestoMA(Ta)
    # do the job
    RH = float(100)*SH*(c.Md/c.Mv)*ps/VPsat(Ta)
    # convert back to ndarray if input is not a masked array
    if WasND: RH, _ = pfp_utils.MAtoSeries(RH)
    return RH

def densitytimesspecificheat(rhow,Cpw,rhoa,Cpa):
    '''
    Product of air density and specific heat capacity for moist air.
    '''
    return rhow*Cpw+rhoa*Cpa

def specificheatcapacitydryair(Tv):
    '''
    Specific heat capacity of air at constant pressure.
    USEAGE:
     cpd = pfp_mf.specificheatcapacitydryair(Tv)
    INPUT:
     Tv - virtual temperature (from sonic anemometer), degC
    OUTPUT:
     cpd - specific heat capacity of dry air at constant pressure, J/kg/K
    SOURCE:
     EddyPro source code
    '''
    cpd = float(1005)+((Tv+23.12)**2)/float(3364)
    return cpd

def specificheatcapacitywatervapour(Ta, AH):
    '''
    Specific heat capacity of water vapour at constant pressure.
    USEAGE:
     cpv = pfp_mf.specificheatcapacitywatervapour(Ta,AH)
    INPUT:
     Ta - air temperature, degC
     AH - absolute humidity, %
    OUTPUT:
     cpv - specific heat capacity of water vapour at constant pressure, J/kg/K
    SOURCE:
     EddyPro source code
    '''
    RH = relativehumidityfromabsolutehumidity(AH,Ta)
    cpv = float(1859)+0.13*RH+(0.193+0.0056*RH)*Ta+(0.001+0.00005*RH)*Ta**2
    return cpv

def specificheatmoistair(SH):
    # Calculate Cp of moist air, from Stull 1988
    #  Cp - specific heat of dry air at constant pressure, J/kg-K
    #  SH - specific humidity
    # Returns
    #  Cpm - specific heat of moist air at constant pressure, J/kg-K
    Cpm = c.Cpd * (1 + 0.84 * SH)
    return Cpm

def specifichumidity(mr):
    # Calculate specific humidity from mixing ratio
    #  mr - mixing ration, kg/kg
    # Returns
    #  SH = specific humidity, kg/kg
    SH = mr/(1+mr)
    return SH

def specifichumidityfromRH(RH, T, p):
    # Specific humidity (kg/kg) from relative humidity, temperature and pressure
    #  RH is the relative humidity, percent
    #  T is the air temperature, degC
    #  p is the atmospheric pressure, kPa
    # Returns
    #  SH = specific humidity, kg/kg
    SH = (c.Mv / c.Md) * (0.01 * RH * VPsat(T) / p)
    return SH

def tafromtv(Tv,SH):
    # Calculate air temperature from virtual temperature using formula
    # from Campbell Scientific CSAT manual.
    # NOTE: this differs from the usual definition by using 0.51 not 0.61
    #  Tv - virtual temperature, degC
    #  SH - specific humidity, kg/kg
    # Returns
    #  Ta - air temperature, degC
    Ta = ((Tv+273.15)/(1+0.51*SH))-273.15
    return Ta

def tvfromta(Ta, mr):
    # Calculate virtual temperature from air temperature using formula
    # from Stull 1988.
    #  Ta - virtual temperature, degC
    #  mr - H2O mixing ratio
    # Returns
    #  Tv - virtual temperature, degC
    Tv = ((Ta + 273.15)*(1 + 0.61*mr)) - 273.15
    return Tv

def theta(T,p):
    # Calculate potential temperature from air temperature and pressure
    #  T - air temperature, degC
    #  p - pressure, kPa
    # Returns
    #  theta - potential temperature, K
    return (T+273.15)*(100/p)**0.286

def vapourpressure(AH,Ta):
    # Calculate vapour pressure from absolute humidity and temperature
    #  AH - absolute humidity, g/m^3
    #  Ta - air temperature, degC
    # Returns
    #  vp - vapour pressure, kPa
    vp = 0.000001*AH*(Ta+273.15)*c.R/c.Mv
    return vp

def virtualtheta(theta_in, mr):
    # Calculate virtual potential temperature
    #  theta - potential temperature, K
    #  mr - mixing ratio, kg/kg
    # Returns
    #  Tvp - virtual potential temperature, K
    Tvp = theta_in * (1 + (0.61 * mr))
    return Tvp



from .units import ureg

beta = 5       # "beta" value in adiabatic correction to wind profile
Cd = ureg('840.0 J/kg/K')      # heat capacity of mineral component of soil, J/kg/K
Co = ureg('1920.0 J/kg/K')     # heat capacity of organic component of soil, J/kg/K
Cp = ureg('1004.67 J/kg/K')    # specific heat of dry air at constant pressure, J/kg-K
Cpd = ureg('1004.67 J/kg/K')   # specific heat of dry air at constant pressure, J/kg-K
Cw = ureg('4180.0 J/kg/K')    # heat capacity of water, J/kg/K
D0 = 10.       # specific humidity deficit threshold for Lasslop et al (2010) NEE expression
E0_long = 100  # long term activation energy, default value
eps = 0.0000001 # a small number for comparing floats
g = 9.81       # gravitation constant
gamma = 28     # "gamma" value in adiabatic correction to wind profile
g2kg = 1E-3    # convert grams to kilograms
k = 0.4        # von Karmans constant
Lv = ureg('2453600 J/kg')   # latent heat of vapourisation, J/kg
Mc = ureg('0.0120107 kg/mol')  # molecular weight of carbon, kg/mol
Mco2 = ureg('0.04401 kg/mol')  # molecular weight of carbon dioxide, kg/mol
Md = ureg('0.02897 kg/mol')   # molecular weight of dry air, kg/mol
missing_value = -9999         # missing data value
large_value = 1E35            # large value
small_value = -1E35           # small value
Mv = ureg('0.01802 kg/mol') # 0.01802   # molecular weight of water vapour, kg/mol
mu = Md/Mv     # ratio of dry air molecular weight to water vapour molecular weight
rho_water = ureg('1000.0 kg/m^3') # density of water, kg/m^3
R = ureg('8.314 J/mol/K')      # universal gas constant, J/mol.K
Rd = ureg('287.04 J/kg/K')    # gas constant for dry air, J/kg/K
Rv = ureg('461.5 J/kg/K')     # gas constant for water vapour, J/kg/K
Pi = 3.14159   # Pi
sb = ureg('5.6704E-8 W/m^2/K^4')  # Stefan-Boltzman constant, W/m^2/K^4
Tref = ureg('15.0 delta_degC')    # reference temperature in the Lloyd-Taylor respiration equation, degC
T0   = ureg('-46.02 delta_degC')  # zero temp[erature in the Lloyd-Taylor respiration equation, degC
P0a = ureg('10**5 Pa')  # reference pressure for potential temperature, Pa
Tb = 1800      # 30-min period, seconds
C2K = 273.15   # convert degrees celsius to kelvin
# dictionary of instrument characteristics
# used in pfp_ts.MassmanStandard(), pfp_compliance.l1_check_sonic_type() and
# pfp_compliance.l1_check_irga_type()
# 'no_sonic' is for IRGAs used in profile measurements
instruments = {"sonics": {"CSAT3": {"lwVert": 0.115, "lwHor": 0.058, "lTv": 0.115},
                          "CSAT3A": {"lwVert": 0.115, "lwHor": 0.058, "lTv": 0.115},
                          "CSAT3B": {"lwVert": 0.115, "lwHor": 0.058, "lTv": 0.115},
                          "WindMaster-Pro": {"lwVert": 0.106, "lwHor": 0.107, "lTv": 0.106}},
               "irgas": {"open_path": {"Li-7500": {"dIRGA": 0.0095, "lIRGA": 0.127},
                                       "Li-7500A": {"dIRGA": 0.0095, "lIRGA": 0.127},
                                       "Li-7500RS": {"dIRGA": 0.0095, "lIRGA": 0.127},
                                       "Li-7500DS": {"dIRGA": 0.0095, "lIRGA": 0.127},
                                       "EC150": {"dIRGA": 0.01, "lIRGA": 0.154},
                                       "IRGASON": {"dIRGA": 0.01, "lIRGA": 0.154}},
                         "closed_path": {"Li-7200": {"dIRGA": 0.0064, "lIRGA": 0.125},
                                         "Li-7200RS": {"dIRGA": 0.0064, "lIRGA": 0.125},
                                         "Li-7200DS": {"dIRGA": 0.0064, "lIRGA": 0.125},
                                         "EC155": {"dIRGA": 0.008, "lIRGA": 0.120},
                                         "Li-830": {"dIRGA": None, "lIRGA": None},
                                         "Li-840": {"dIRGA": None, "lIRGA": None},
                                         "Li-850": {"dIRGA": None, "lIRGA": None},
                                         "None": {"dIRGA": None, "lIRGA": None}}}}
# dictionary of site names and time zones
tz_dict = {"adelaideriver":"Australia/Darwin",
           "alicespringsmulga":"Australia/Darwin",
           "arcturus":"Australia/Brisbane",
           "calperum":"Australia/Adelaide",
           "capetribulation":"Australia/Brisbane",
           "cowbay":"Australia/Brisbane",
           "cumberlandplains":"Australia/Sydney",
           "cup_ec":"Australia/Sydney",
           "daintree":"Australia/Brisbane",
           "dalypasture":"Australia/Darwin",
           "dalyregrowth":"Australia/Darwin",
           "dalyuncleared":"Australia/Darwin",
           "dargo":"Australia/Melbourne",
           "dryriver":"Australia/Darwin",
           "foggdam":"Australia/Darwin",
           "gingin":"Australia/Perth",
           "greatwestern":"Australia/Perth",
           "gww":"Australia/Perth",
           "howardsprings":"Australia/Darwin",
           "litchfield":"Australia/Darwin",
           "nimmo":"Australia/Sydney",
           "reddirt":"Australia/Darwin",
           "riggs":"Australia/Melbourne",
           "robson":"Australia/Brisbane",
           "samford":"Australia/Brisbane",
           "sturtplains":"Australia/Darwin",
           "titreeeast":"Australia/Darwin",
           "tumbarumba":"Australia/Canberra",
           "wallaby":"Australia/Melbourne",
           "warra":"Australia/Hobart",
           "whroo":"Australia/Melbourne",
           "wombat":"Australia/Melbourne",
           "yanco_jaxa":"Australia/Sydney"}
units_synonyms = {"Fsd":["W/m^2","W+1m-2"],
                  "Fsu":["W/m^2","W+1m-2"],
                  "Fld":["W/m^2","W+1m-2"],
                  "Flu":["W/m^2","W+1m-2"],
                  "Fn":["W/m^2","W+1m-2"],
                  "Fg":["W/m^2","W+1m-2"],
                  "Precip":["mm"],
                  "ps":["kPa"],
                  "RH":["%","percent"],
                  "Sws":["frac","m^3/m^3","m+3m-3"],
                  "Ta":["C","degC"],
                  "Ts":["C","degC"],
                  "Wd":["degT","deg","degrees"],
                  "Ws":["m/s","m+1s-1"]}


import numpy as np


def time_lag(length=711, diameter=5.3, pump=15, dt=20):
    # default time lag in number of data points
    return int(np.round((length * (np.pi * (diameter/2)**2) * (10**-6) / pump) * 60 * dt))



import os
import logging
import pandas as pd
import glob
from functools import reduce
from . import import_ascii, import_binary, read_data

logger = logging.getLogger()


def import_data(setup, datetimes, file_importer, **raw_kwargs):
    setup_raw_kw = setup.get('Files', {})

    head_nlines = int(setup_raw_kw['in_firstdatarow'])-2
    header_col = int(setup_raw_kw['in_headerrow'])-1
    sep = setup_raw_kw['in_separator']
    na_values = [-9999]
    glob_files = setup_raw_kw['glob_files']

    def extract_datetime_parts(dt_input):
        # Parse if input is string
        dt = pd.to_datetime(dt_input)
        return dict(
            year=dt.strftime('%Y'),
            yr=dt.strftime('%y'),
            month=dt.strftime('%m'),
            day=dt.strftime('%d'),
            hour=dt.strftime('%H'),
            minute=dt.strftime('%M'),
        )

    def select_files_based_on_datetime(pattern, dt_input):
        datetime_parts = extract_datetime_parts(dt_input)
        return glob.glob(pattern.format(**datetime_parts))

    files = [f for dt in datetimes for f in select_files_based_on_datetime(
        glob_files, dt)]

    if not files:
        if glob.glob(os.path.dirname(glob_files)):
            logger.warning(
                f'No files found based on dates. Looked in `{os.path.dirname(glob_files)}`.')
        else:
            logger.warning(
                f'Folder `{os.path.dirname(glob_files)}` is empty.')
        
        return pd.DataFrame()

    raw_data = reduce(lambda left, right: pd.concat([left, right], ignore_index=True),
                      [file_importer(
                          f, head_nlines=head_nlines, header_col=header_col, sep=sep, na_values=na_values) for f in files])

    raw_data = read_data.format_column_names(raw_data)

    raw_data = read_data.resolve_timestamp(
        raw_data, 'timestamp')

    raw_data = raw_data.sort_values(by='timestamp')

    return raw_data





import xarray as xr
import numpy as np
import pandas as pd

def e(ds):
    if 'zm' in ds:
        zm = ds['zm']
    else:
        zm = ds.attrs.get('zm', None)

    if 'z0' in ds:
        z0 = ds['z0']
    else:
        z0 = ds.attrs.get('z0', None)

    if 'zL' in ds:
        zL = ds['zL']
    else:
        zL = (zm-z0) / ds['ol']

def integral_turbulence_characteristics(ds, x1="w", x2="co2", 
                                        zL=1, latitude=None, **kwargs):
    '''
    Add turbulence flag to dataset.
    '''
    if not latitude:
        latitude = ds.attrs.get('latitude', None)
        
    # n = {k: k for k in ['zL', 'zm', 'z0', 'ol', 'ustar', 'sigmaw']}
    # n.update(**{k: v for k, v in kwargs.items() if k in n.keys()})

    assert len(list(ds.dims)) == 1, 'Data must be a single time series.'

    sTp_Tpstar_mo = 2 * np.abs(zL)**(1/8)

    if (np.abs(zL) < 0.032):
                 sTp_Tpstar_mo = 1.3
    if (np.abs(zL) > -0.2) and (np.abs(zL) < 0.4):
        # f: Coriolis parameter
        f = 2. * 2.*np.pi/(24.*60.*60.) * np.sin(np.deg2rad(latitude))
        sTp_Tpstar_mo = 0.21 * np.log(1. * f / ds['ustar']) + 3.1
    
    sw_ustar = ds['std_w'] / ds['ustar']

    stat = np.floor(
        np.abs((sTp_Tpstar_mo - sw_ustar) / sTp_Tpstar_mo) * 100.)
    if (sTp_Tpstar_mo <= 10**-6):
        stat = 0

    flag = 2
    if stat <= 100:
        flag = 1
    if stat <= 30:
        flag = 0

    ds = ds.assign({f'qaqc_itc_{x1}_{x2}': stat,
                    f'qaqc_itc_flag_{x1}_{x2}': flag})

    # for each group calculate the cov
    # calculate the mean of the cov for all groups
    # compare the mean with the cov
    # return % and flag (0,1,2)

    return ds

    return type(
        'var_', (object,),
        {"test": stat,
         "flag": flag,
         "meta": {}}
    )


import xarray as xr
from . import wilczak_et_al_2001

available_corrections = {
    'double_rotation':
    type('var_', (object,), {'run': wilczak_et_al_2001.double_rotation,
         'name': 'double rotation (Wilczak, et al., 2001)'}),
    'triple_rotation': 
    type('var_', (object,), {'run': wilczak_et_al_2001.triple_rotation,
         'name': 'triple rotation (Wilczak, et al., 2001)'}),
    'planarfit': 
    type('var_', (object,), {'run': wilczak_et_al_2001.planarfit,
         'name': 'planar fit (Wilczak, et al., 2001)'}),
}

def apply_axis_rotation(data):
     assert all([v in data for v in ['u', 'v', 'w']]), 'Missing required variables in axis rotation.' 

     correction = available_corrections.get(
         data.attrs.get('Corrections', {}).get('axis_rotation', {}).get('method', None), None)

     data = data.assign(u_unrot=data.u, v_unrot=data.v, w_unrot=data.w)

     if correction:
          dr = correction.run(data.u, data.v, data.w)
          dr = vars(dr)

          for k in ['u', 'v', 'w']:
               data[k].data = dr[k]
               
               data[k] = data[k].assign_attrs(axis_rotation_method=correction.name)

               if dr.get('meta', None):
                   data[k] = data[k].assign_attrs(
                       time_lag_params=dr.get('meta', ''))
               
          for k, v in dr.items():
               if k.startswith('__') or (k in ['u', 'v', 'w', 'meta']):
                    continue
               else:
                   var_name = f'axis_rotation_params_{k}'
                   data = data.assign(**{var_name: v})
                   data[var_name] = data[var_name].assign_attrs(
                         time_lag_method=correction.name)
          # if isinstance(
          #      dr.u.data, xr.DataArray) else dr.u
          # data['v'].data = dr.v if isinstance(
          #      dr.v.data, xr.DataArray) else dr.v
          # data['w'].data = dr.w if isinstance(
          #      dr.w.data, xr.DataArray) else dr.w

          # data['u'] = data['u'].assign_attrs(
          #      **dict(axis_rotation_method=correction.name,
          #                axis_rotation_params={k: v for k, v in dr.items() if not k.startswith('__') and k not in ['u', 'v', 'w']}))

          data.attrs['corrections_applied'] += f'axis rotation w/ {correction.name}; '
     return data



import xarray as xr
import numpy as np
import pandas as pd


def stationarity_test(ds, x1="w", x2="co2", n=6):
    '''
    Add stationarity flag to dataset.
    
    Parameters
    ----------
    flagdata: pandas DataFrame with stationarity flags
    data2: pandas DataFrame with complementary data for comparison

    flagdata: data to include stationarity flag
    data2: complementary data at a faster pace for comparison
    cov_name: variable used for comparison
    one_dta_per: input how many data2 observations are accounted per flagdata observation
    '''
    
    # assert len(list(ds.dims)) == 1, 'Data must be a single time series.'

    # Check if x1 and x2 are in the dataset
    if x1 not in ds or x2 not in ds:
        raise ValueError(f"Variables '{x1}' and '{x2}' must be present in the dataset.")
    
    dim = list(ds.dims)[0]

    sum_ = []
    chunk_size = ds.sizes[dim] // n
    for i in range(n):
        start = i * chunk_size
        end = (i + 1) * chunk_size if i < n - 1 else ds.sizes[dim]
        ds_ = ds.isel({dim: slice(start, end)})
        sum_ += [xr.cov(ds_[x1], ds_[x2])]
    sum_ = np.nanmean(sum_)

    cov_ = xr.cov(ds[x1], ds[x2])

    stat = np.floor(
        np.abs((sum_ - cov_) / cov_) * 100.)
    if (sum_ <= -9000.) or (cov_ <= -9000.) or (cov_ > 99999):
        stat = 99999

    flag = 2
    if stat <= 100:
        flag = 1
    if stat <= 30:
        flag = 0

    ds = ds.assign({f'qaqc_stationarity_{x1}_{x2}': stat,
                    f'qaqc_stationarity_flag_{x1}_{x2}': flag})

    # for each group calculate the cov
    # calculate the mean of the cov for all groups
    # compare the mean with the cov
    # return % and flag (0,1,2)

    return ds

    return type(
        'var_', (object,),
        {"test": stat,
         "flag": flag,
         "meta": {}}
    )

"""
ADD FLAGS
"""


def fITC(flagdata, latitude, **kwargs):
    n = {k: k for k in ['zL', 'zm', 'z0', 'ol', 'ustar', 'sigmaw']}
    n.update(**{k: v for k, v in kwargs.items() if k in n.keys()})

    if n['zL'] not in flagdata.columns:
        flagdata[n['zL']] = (flagdata[n['zm']]-flagdata[n['z0']]) / flagdata.ol

    flagdata['sTp_Tpstar_mo'] = 2 * np.abs(flagdata[n['zL']])**(1/8)

    flagdata.loc[(np.abs(flagdata[n['zL']]) < 0.032), 'sTp_Tpstar_mo'] = 1.3

    # f: Coriolis parameter
    f = 2. * 2.*np.pi/(24.*60.*60.) * np.sin(np.deg2rad(latitude))

    cond_ = (np.abs(flagdata[n['zL']]) > -0.2) * \
        (np.abs(flagdata[n['zL']]) < 0.4)
    flagdata.loc[cond_, 'sTp_Tpstar_mo'] = (
        0.21 * np.log(1. * f / flagdata[n['ustar']]) + 3.1)[cond_]

    flagdata['sw_ustar'] = flagdata[n['sigmaw']] / flagdata[n['ustar']]
    flagdata['itc_co2'] = np.floor(np.abs((flagdata.sTp_Tpstar_mo -
                                           flagdata.sw_ustar) / flagdata.sTp_Tpstar_mo) * 100)

    flagdata.loc[flagdata.sTp_Tpstar_mo < 10**-6, 'itc_co2'] = 0

    flagdata['fITC'] = 2
    flagdata.loc[flagdata.itc_co2 <= 100, 'fITC'] = 1
    flagdata.loc[flagdata.itc_co2 <= 30, 'fITC'] = 0

    return flagdata


def fSTA(flagdata, data2, cov_name='cov_wco2', bydate=False, one_dta_per=None):
    '''
    Add stationarity flag to dataset.

    flagdata: data to include stationarity flag
    data2: complementary data at a faster pace for comparison
    cov_name: variable used for comparison
    one_dta_per: input how many data2 observations are accounted per flagdata observation
    '''

    if bydate:
        """
        breaks = np.unique(data2[bydate].dt.ceil(
            '30min'), return_index=True)[1]
        breaks.sort()
        #breaks = breaks[1:] if breaks[0] == 0 else breaks
        
        flagdata['sum_wco2'] = np.array([np.nanmean(p) for p in np.split(
            data2[cov_name], breaks, axis=0)]).ravel()
        """
        data2[bydate] = data2[bydate].dt.ceil('30min')
        data2 = data2.groupby(bydate)[cov_name].apply(np.nanmean)
        data2 = data2.reset_index()
        data2.columns = [bydate, 'sum_wco2']
        flagdata = pd.merge(flagdata, data2, 'left', bydate)
        # flagdata['sum_wco2'] = np.array(data2.groupby("_group_")[cov_name].apply(np.nanmean)).ravel()

    else:
        if one_dta_per:
            flagdata['sum_wco2'] = np.nanmean(np.array(
                data2[cov_name]).reshape(-1, int(one_dta_per)), axis=1).ravel()[:len(flagdata)]
        else:
            flagdata['sum_wco2'] = np.nanmean(np.array(
                data2[cov_name]).reshape(-1, int(round(len(data2)/len(flagdata)))), axis=1).ravel()

    flagdata['stat_wco2'] = np.floor(
        np.abs((flagdata.sum_wco2 - flagdata[cov_name]) / flagdata[cov_name]) * 100.)

    flagdata.loc[flagdata['sum_wco2'] <= -9000., ['stat_wco2']] = 99999
    flagdata.loc[flagdata[cov_name] <= -9000., ['stat_wco2']] = 99999
    flagdata.loc[flagdata[cov_name]
                 > 99999, ['stat_wco2']] = 99999

    flagdata['fSTA'] = 2
    flagdata.loc[flagdata.stat_wco2 <= 100, 'fSTA'] = 1
    flagdata.loc[flagdata.stat_wco2 <= 30, 'fSTA'] = 0
    return flagdata



"""
Simply read data from a file and return a pandas DataFrame.
This module provides a function to read data from a file and return it as a pandas DataFrame.

It supports various file formats including CSV, Excel, and NetCDF.
"""
import xarray as xr
import re
import pandas as pd
import numpy as np
import datetime
from pandas.api.types import is_numeric_dtype, is_object_dtype
from ..._core.units import resolve_unit

# Read eddypro
# Data to netcdf4 with a format that allows :
# import xarray as xr
# import numpy as np

# # Create example data
# avgtime = pd.date_range('2023-01-01', periods=10)
# rawtime = pd.date_range('2023-01-01', periods=10)
# lat = np.arange(10)
# lon = np.arange(10)
# temperature = np.random.rand(len(time), len(lat), len(lon))
# humidity = np.random.rand(len(time), len(lat), len(lon))

# # Create an xarray Dataset
# ds = xr.Dataset(
#     {
#         'temperature': (('avgtime', 'lat', 'lon'), temperature),
#         'humidity': (('rawtime', 'lat', 'lon'), humidity),
#     },
#     coords={
#         'time': time,
#         'lat': lat,
#         'lon': lon,
#     }
# )

# # Access variables like columns
# temperature_data = ds['temperature']
# humidity_data = ds['humidity']

# # Perform operations on the data
# mean_temperature = temperature_data.mean(dim='time')


def d0_d1_from_time_range(trange):
    if trange is None:
        return None, None
    d0 = pd.to_datetime(trange.split(
        '-', 1)[0]) if trange.split('-', 1)[0] else None
    d1 = pd.to_datetime(trange.split(
        '-', 1)[1]) if trange.split('-', 1)[1] else None
    return d0, d1


def generate_time_windows(tmin, tmax, fastfreq, slowfreq, include='both'):
    tmin = pd.to_datetime(tmin)
    tmax = pd.to_datetime(tmax)
    slow_range = pd.date_range(tmin, tmax, freq=slowfreq).floor(slowfreq)
    slow_delta = pd.Timedelta(slowfreq)

    result = []
    for p in slow_range:
        start = max(p, tmin)
        end = min(p + slow_delta, tmax)
        times = pd.date_range(start, end, freq=fastfreq)

        if include == 'left':
            times = times[:-1]
        elif include == 'right':
            times = times[1:]
        elif include != 'both':
            continue  # or raise ValueError("Invalid 'include' value")

        result.append(times)

    return result

def resolve_timestamp(data, timestamp_name='timestamp'):
    # Create a lowercase mapping of column names to original column names
    col_map = {col.lower(): col for col in data.columns}

    def has_cols(*cols):
        return all(col.lower() in col_map for col in cols)

    if timestamp_name.lower() in col_map:
        try:
            data[timestamp_name] = pd.to_datetime(
                data[col_map['timestamp']].astype(
                    str), format='%Y%m%d%H%M%S.%f'
            )
        except ValueError:
            data[timestamp_name] = pd.to_datetime(
                data[col_map['timestamp']], errors='coerce'
            )
    elif 'timestamp_end' in col_map:
        data[timestamp_name] = pd.to_datetime(
            data[col_map['timestamp_end']], format='%Y%m%d%H%M'
        )
    elif has_cols('date', 'time'):
        data[timestamp_name] = pd.to_datetime(
            data[col_map['date']] + " " + data[col_map['time']]
        )
    elif has_cols('seconds', 'nanoseconds'):
        data[timestamp_name] = pd.to_datetime(
            data[col_map['seconds']] + data[col_map['nanoseconds']] * 10**-9, unit='s'
        )
    elif has_cols('year', 'month', 'day', 'hour', 'minute'):
        data[timestamp_name] = pd.to_datetime(
            data[col_map['year']].astype(str).str.zfill(4) + '-' +
            data[col_map['month']].astype(str).str.zfill(2) + '-' +
            data[col_map['day']].astype(str).str.zfill(2) + 'T' +
            data[col_map['hour']].astype(str).str.zfill(2) +
            data[col_map['minute']].astype(str).str.zfill(2)
        )

    return data



def extract_timestamp(df, tname, datestr, dt, date_format, datefomatfrom, datefomatto):
    """
    Extract timestamp
    """
    if tname not in df.columns or datefomatfrom == 'drop':
        if "date" in df.columns and "time" in df.columns:
            df[tname] = pd.to_datetime(
                df.date + " " + df.time, format='%Y-%m-%d %H:%M')
        else:
            df[tname] = pd.to_datetime(
                datestr, format=date_format) - datetime.timedelta(seconds=dt) * (len(df)-1 + -1*df.index)
            # td, format=date_format) + datetime.timedelta(seconds=dt) * (df_td.index)
            df[tname] = df[tname].dt.strftime(
                datefomatto)
    else:
        try:
            if is_numeric_dtype(df[tname]):
                df.loc[:, tname] = df.loc[:, tname].apply(lambda e: pd.to_datetime(
                    '%.2f' % e, format=datefomatfrom).strftime(datefomatto))
            elif is_object_dtype(df[tname]):
                df.loc[:, tname] = df.loc[:, tname].apply(
                    lambda e: pd.to_datetime(e).strftime(datefomatto))
            else:
                df.loc[:, tname] = pd.to_datetime(
                    df[tname], format=datefomatfrom).strftime(datefomatto)
        except:
            # warnings.warn(f'error when converting {tname} from {datefomatfrom} to {datefomatto}.')
            pass
    return df


def format_column_names(data):
    data.columns = [re.sub(r'\s+', '_', str(c).strip().lower()) for c in data.columns]

    # Variable mapping
    rename_map = {
        # 'h2o': 'h2o_c', 'h2o_dry': 'h2o',
        'co2': 'co2_c', 'co2_dry': 'co2',
        't_cell_out': 'ta',
    }

    # Filter only the variables available in the dataset
    rename_map_filtered = {k: v for k,
                           v in rename_map.items() if k in data.columns}

    # Rename variables to standard names
    data = data.rename(columns=rename_map_filtered)

    return data


def init_data(time_coord='timestamp', **attrs):
    data = pd.DataFrame()

    # Create an xarray Dataset
    global_attrs = attrs.copy()
    global_attrs.update(global_attrs.pop('Global', {}))

    vars_attrs = {k: global_attrs.pop(k)
                  for k, v in attrs.items() if k in data.columns}
    vars_attrs.update(global_attrs.pop('Variables', {}))

    data.insert(0, 'timestamp_end', pd.NaT)
    data.insert(0, 'timestamp_start', pd.NaT)

    data.insert(0, 'date', pd.NaT)
    data.insert(0, 'time', pd.NA)
    data.insert(0, 'latitude', data['latitude']
                if 'latitude' in data else attrs.get('latitude', None))
    data.insert(0, 'longitude', data['longitude']
                if 'longitude' in data else attrs.get('longitude', None))

    ds = (data
          .dropna(subset=['date', 'time', 'latitude', 'longitude'], how='any')
          .set_index(['date', 'time', 'latitude', 'longitude'])
          .to_xarray())

    ds = ds.assign_attrs(**global_attrs)
    ds = ds.assign_attrs(Variables=vars_attrs)

    return ds


def format_data(data, time_coord='timestamp', avg='30Min', **attrs):
    # Create an xarray Dataset
    global_attrs = attrs.copy()
    global_attrs.update(global_attrs.pop('Global', {}))

    vars_attrs = {k: global_attrs.pop(k)
                  for k, v in attrs.items() if k in data.columns}
    vars_attrs.update(global_attrs.pop('Variables', {}))

    timestamp_end = data[time_coord].dt.ceil(avg)
    timestamp_start = timestamp_end - pd.Timedelta(avg)

    data.insert(0, 'date', timestamp_end)
    data.insert(0, 'time', (data[time_coord] -
                            timestamp_start) / pd.Timedelta('1s'))
    data.insert(0, 'latitude', data['latitude']
                if 'latitude' in data else attrs.get('latitude', None))
    data.insert(0, 'longitude', data['longitude']
                if 'longitude' in data else attrs.get('longitude', None))
    
    # Drop rows with missing values in the index columns
    data = data.dropna(subset=['date', 'time', 'latitude', 'longitude'], how='any')

    # Set the multi-index
    data = data.set_index(['date', 'time', 'latitude', 'longitude'])

    # Convert to xarray Dataset
    ds = data.to_xarray()

    ds['date'].attrs.update(**{'value': 'timestamp_end', 'delta': pd.Timedelta(avg)})

    # ds = (data
    #       .dropna(subset=['date', 'time', 'latitude', 'longitude'], how='any')
    #       .set_index(['date', 'time', 'latitude', 'longitude'])
    #       .to_xarray())

    # ds = data.set_index([time_coord]).to_xarray()
    # ds = ds.assign_coords(date=ds['date'], time=ds['time'])
    # ds = ds.set_index(timestamp=["date", "time"]).unstack("timestamp")

    ds = ds.assign_attrs(**global_attrs)

    for k, v in vars_attrs.items():
        if k in ds:
            ds[k] = ds[k].assign_attrs(**v)

        elif v.get('csv', None):
            if v['csv'].get('name', None) in ds:
                ds[k] = ds[k].assign_attrs(**v['csv'].get('Attr', {}))
            elif v['csv'].get('column', None):
                var_col_nb = int(float(v['csv']['column'])) - 1
                # var_col_nb += 2
                var_col = data.columns[var_col_nb]
                ds[var_col] = ds[var_col].assign_attrs(
                    **v.get('Attr', {}))

    # Make unit aware
    for var in ds.data_vars:
        units = ds[var].attrs.get(
            'units', ds[var].attrs.get('unit_in', 'dimensionless'))
        try:
            # keep_attrs = ds[var].attrs.copy()
            resolved_unit = resolve_unit(units)
            ds[var].data = ds[var] * resolved_unit
            ds[var].attrs.update({'units': resolved_unit})
            # keep_attrs.update({'units': resolved_unit})
            # ds[var].attrs = keep_attrs
        except Exception as e:
            ds[var].attrs.update({'units': 'dimensionless'})
            # logger.debug(f"Error for variable `{var}`: {e}")
            pass
    return ds



import xarray as xr
import numpy as np
import matplotlib.pyplot as plt

def fft_over_time(da):
    # Interpolate missing data along time
    da_interp = da.interpolate_na(dim='time', fill_value="extrapolate")

    # Apply FFT using apply_ufunc to preserve xarray metadata
    return xr.apply_ufunc(
        np.fft.fft,
        da_interp,
        input_core_dims=[["time"]],
        output_core_dims=[["frequency"]],
        dask="parallelized",
        output_dtypes=[np.complex128]
    )


def cospectrum(*args):
    return xr.apply_ufunc(
        lambda a, b: np.real(a * np.conj(b)),
        *args,
        input_core_dims=[["frequency"], ["frequency"]],
        output_core_dims=[["frequency"]],
        dask="parallelized", output_dtypes=[np.float64]
    )


def get_spectra(ds, variables=['u', 'v', 'w', 'co2', 'h2o', 't_sonic']):
    # Frequency axis
    freq = np.fft.fftfreq(len(ds.time))

    ds = ds.assign_coords(frequency=freq)

    for v in variables:
        if v in ds:
            ds[f'spectrum_{v}'] = ds.groupby(['date', 'latitude', 'longitude']).apply(
                lambda ds: fft_over_time(ds[v]))
    
    return ds


def get_cospectra(ds, variables=[['u', 'v'], ['w', 'co2'], ['w', 'h2o'], ['w', 't_sonic']]):
    # Frequency axis
    freq = np.fft.fftfreq(len(ds.time))

    ds = ds.assign_coords(frequency=freq)

    for x1, x2 in variables:
        if f'spectrum_{x1}' in ds and f'spectrum_{x2}' in ds:
            ds[f'cospectrum_{x1}_{x2}'] = ds.groupby(['date', 'latitude', 'longitude']).apply(
                lambda ds: cospectrum(ds[f'spectrum_{x1}'], ds[f'spectrum_{x2}']))
    return ds

# data['cospectrum_wts'] = data.groupby(['date', 'latitude', 'longitude']).apply(
#     lambda ds: cospectrum(ds.spectrum_w, ds.spectrum_ts))

def plot(ds):
    # Plot only the positive half (since it's symmetric)
    half = len(ds.frequency) // 2
    freq = ds.frequency

    power = (ds['cospectrum_wts'] / ds['cov_w_t_sonic']).pipe(abs).mean(
        dim=['date', 'latitude', 'longitude'])
    plt.plot(freq[:half], power[:half], c='r')

    power = (ds['cospectrum_wco2'] / ds['cov_w_co2']).pipe(abs).mean(
        dim=['date', 'latitude', 'longitude'])
    plt.plot(freq[:half], power[:half], c='k')

    plt.title("Power Spectrum")
    plt.xlabel("Frequency")
    plt.ylabel("Power (|FFT|²)")
    plt.grid(True)
    plt.yscale('log')
    plt.xscale('log')
    plt.show()



import logging
import pandas as pd
import xarray as xr
from . import version, _core, io, process, corrections, qaqc, external, compatibility
from ._core.units import convert_to_prefered_units
from concurrent.futures import ProcessPoolExecutor

logger = logging.getLogger()





# standard modules
# from ux import opf_processing
from gui import pfp_threading
from gui import pfp_log
# from ux import pfp_io
# from ux import pfp_gui
# from ux import pfp_compliance
from version import version_name, version_number
from PyQt5.QtCore import *
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
import datetime
import faulthandler
import logging
import os
import sys
import traceback
import warnings
import webbrowser
# 3rd party modules
from configobj import ConfigObj
import netCDF4
import matplotlib
matplotlib.use("QT5Agg")
# Local modules
import version
from  _core import utils
from gui import top_level
from gui import gui
# sys.path.insert(0, 'scripts')
# PFP new modules

faulthandler.enable()
warnings.filterwarnings("ignore", category=Warning)

# now check the logfiles and plots directories are present
base_path = utils.get_base_path()
logfiles_path = os.path.join(base_path, "logfiles", "")
plots_path = os.path.join(base_path, "plots", "")
dir_list = [logfiles_path, plots_path]
for item in dir_list:
    if not os.path.exists(item):
        os.makedirs(item)

now = datetime.datetime.now()
logger_name = f'{version.__name__}_log'
log_file_name = version.__name__ + "_" + now.strftime("%Y%m%d%H%M") + ".log"
log_file_name = os.path.join(logfiles_path, log_file_name)
logger = pfp_log.CreateLogger(logger_name, log_file_name=log_file_name)


class pfp_main_ui(QWidget):
    def __init__(self, pfp_version, textBox):
        super(pfp_main_ui, self).__init__()
        # set the mode attribute
        self.mode = "interactive"
        # menu bar
        self.menubar = QMenuBar(self)
        # File menu
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setTitle("File")
        # File/Convert submenu
        self.menuFileConvert = QMenu(self.menuFile)
        self.menuFileConvert.setTitle("Convert")
        # Edit menu
        self.menuEdit = QMenu(self.menubar)
        self.menuEdit.setTitle("Edit")
        # Run menu
        self.menuRun = QMenu(self.menubar)
        self.menuRun.setTitle("Run")
        # Help menu
        self.menuHelp = QMenu(self.menubar)
        self.menuHelp.setTitle("Help")
        # File menu items: menu actions for control files
        self.actionFileOpen = QAction(self)
        self.actionFileOpen.setText("Open")
        self.actionFileOpen.setShortcut('Ctrl+O')
        self.actionFileSave = QAction(self)
        self.actionFileSave.setText("Save")
        self.actionFileSave.setShortcut('Ctrl+S')
        self.actionFileSaveAs = QAction(self)
        self.actionFileSaveAs.setText("Save As...")
        self.actionFileSaveAs.setShortcut('Shift+Ctrl+S')
        # File/Convert submenu
        self.actionFileConverteddypro2nc = QAction(self)
        self.actionFileConverteddypro2nc.setText("EddyPro to nc")
        self.actionFileConvertnc2biomet = QAction(self)
        self.actionFileConvertnc2biomet.setText("nc to Biomet")
        self.actionFileConvertnc2xls = QAction(self)
        self.actionFileConvertnc2xls.setText("nc to Excel")
        self.actionFileConvertnc2reddyproc = QAction(self)
        self.actionFileConvertnc2reddyproc.setText("nc to REddyProc")
        # File menu item: split netCDF
        self.actionFileSplit = QAction(self)
        self.actionFileSplit.setText("Split")
        self.actionFileSplit.setShortcut('Ctrl+P')
        # File menu item: Quit
        self.actionFileQuit = QAction(self)
        self.actionFileQuit.setText("Quit")
        # the Vinod mod ...
        self.actionFileQuit.setShortcut('Ctrl+Q')
        # Edit menu items
        self.actionEditPreferences = QAction(self)
        self.actionEditPreferences.setText("Preferences...")
        # Run menu items
        self.actionRunCurrent = QAction(self)
        self.actionRunCurrent.setText("Current...")
        self.actionRunCurrent.setShortcut('Ctrl+R')
        self.actionRunClearLogWindow = QAction(self)
        self.actionRunClearLogWindow.setText("Clear log window")
        self.actionStopCurrent = QAction(self)
        self.actionStopCurrent.setText("Stop run")
        # Help menu
        self.actionHelpWiki = QAction(self)
        self.actionHelpWiki.setText("Wiki")
        self.actionHelpAbout = QAction(self)
        self.actionHelpAbout.setText("About")
        # add the actions to the menus
        # File/Convert submenu
        self.menuFileConvert.addAction(self.actionFileConverteddypro2nc)
        self.menuFileConvert.addAction(self.actionFileConvertnc2xls)
        self.menuFileConvert.addAction(self.actionFileConvertnc2biomet)
        self.menuFileConvert.addAction(self.actionFileConvertnc2reddyproc)
        # File menu
        self.menuFile.addAction(self.actionFileOpen)
        self.menuFile.addAction(self.actionFileSave)
        self.menuFile.addAction(self.actionFileSaveAs)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.menuFileConvert.menuAction())
        self.menuFile.addAction(self.actionFileSplit)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionFileQuit)
        # Edit menu
        self.menuEdit.addAction(self.actionEditPreferences)
        # Run menu
        self.menuRun.addAction(self.actionRunCurrent)
        self.menuRun.addAction(self.actionRunClearLogWindow)
        # add individual menus to menu bar
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuRun.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        # Help menu
        self.menuHelp.addAction(self.actionHelpWiki)
        self.menuHelp.addAction(self.actionHelpAbout)

        # create a tab bar
        self.tabs = QTabWidget(self)
        self.tabs.tab_index_all = 0
        self.tabs.tab_index_current = 0
        self.tabs.tab_dict = {}
        # make the tabs closeable
        self.tabs.setTabsClosable(True)
        self.tabs.tabCloseRequested.connect(self.closeTab)
        # add the text editor to the first tab
        self.tabs.addTab(textBox, "Log")
        self.tabs.tab_index_all = self.tabs.tab_index_all + 1
        # hide the tab close icon for the console tab
        self.tabs.tabBar().setTabButton(0, QTabBar.RightSide, None)
        # connect the tab-in-focus signal to the appropriate slot
        self.tabs.currentChanged[int].connect(self.tabSelected)

        # use VBoxLayout to position widgets so they resize with main window
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        # add widgets to the layout
        layout.addWidget(self.menubar)
        layout.addWidget(self.tabs)
        self.setLayout(layout)
        self.setGeometry(50, 50, 800, 600)
        self.setWindowTitle(pfp_version)

        self.threadpool = QThreadPool()

        # Connect signals to slots
        # File menu actions
        self.actionFileConverteddypro2nc.triggered.connect(
            lambda: top_level.do_file_convert_eddypro2setup(None, mode="standard"))
        # self.actionFileConvertnc2biomet.triggered.connect(
        #     lambda: pfp_top_level.do_file_convert_nc2biomet(None, mode="standard"))
        # self.actionFileConvertnc2xls.triggered.connect(
        #     pfp_top_level.do_file_convert_nc2xls)
        # self.actionFileConvertnc2reddyproc.triggered.connect(
        #     lambda: pfp_top_level.do_file_convert_nc2reddyproc(None, mode="standard"))
        self.actionFileOpen.triggered.connect(self.file_open)
        self.actionFileSave.triggered.connect(self.file_save)
        self.actionFileSaveAs.triggered.connect(self.file_save_as)
        # self.actionFileSplit.triggered.connect(pfp_top_level.do_file_split)
        self.actionFileQuit.triggered.connect(QApplication.quit)
        # Edit menu actions
        self.actionEditPreferences.triggered.connect(self.edit_preferences)
        # Run menu actions
        self.actionRunCurrent.triggered.connect(self.run_current)
        self.actionRunClearLogWindow.triggered.connect(
            self.run_clear_log_window)
        # Help menu actions
        self.actionHelpWiki.triggered.connect(self.help_wiki)
        self.actionHelpAbout.triggered.connect(self.help_about)
        # # add the L4 GUI
        # self.l4_ui = pfp_gui.pfp_l4_ui(self)
        # # add the L5 GUI
        # self.solo_gui = pfp_gui.solo_gui(self)

    def file_open(self, file_uri=None):
        """
        Purpose:
         Get a file name, figure out if it is a control file or a netCDF
         file and try to open it.
        """
        # local logger
        logger = logging.getLogger(f"{version.__name__}_log")
        # get the control file path
        if not file_uri:
            file_uri = QFileDialog.getOpenFileName(
                caption="Choose a file ...")[0]
            # check to see if file open was cancelled
            if len(str(file_uri)) == 0:
                return
        self.file_uri = str(file_uri)
        # read the contents of the control file
        logger.info(" Opening " + self.file_uri)
        # check to see if it is a control file
        file_open_success = False
        if not file_open_success:
            try:
                self.file = ConfigObj(self.file_uri, indent_type="    ", list_values=False,
                                      write_empty_values=True)
                file_open_success = True
            except Exception:
                # trying to open a netCDF file will throw UnicodeDecodeError
                # opening a ConfigObj file with syntax erros throws ConfigObjError
                exc_type, _, _ = sys.exc_info()
                if exc_type.__name__ == "UnicodeDecodeError":
                    # most likely a netCDF file
                    pass
                elif exc_type.__name__ == "ConfigObjError":
                    # syntax error in control file
                    msg = "Syntax error in control file, see below for line number"
                    logger.error(msg)
                    error_message = traceback.format_exc()
                    logger.error(error_message)
                    return
                else:
                    # unrecognised file type
                    msg = "File must be either a control file or a netCDF file"
                    logger.error(msg)
                    error_message = traceback.format_exc()
                    logger.error(error_message)
                    return
        # check to see if it is a netCDF file
        if not file_open_success:
            try:
                self.file = netCDF4.Dataset(self.file_uri, "r")
                file_open_success = True
            except:
                # unrecognised file type
                msg = "File must be either a control file or a netCDF file"
                logger.error(msg)
                error_message = traceback.format_exc()
                logger.error(error_message)
                return
        # do the business depending on the file type
        if isinstance(self.file, ConfigObj):
            # we are opening a control file
            self.open_control_file()
        elif isinstance(self.file, netCDF4._netCDF4.Dataset):
            # we are opening a netCDF file
            self.open_netcdf_file()
        else:
            # unrecognised file type
            msg = "File must be either a control file or a netCDF file"
            logger.error(msg)
            error_message = traceback.format_exc()
            logger.error(error_message)
        return

    def help_about(self):
        msg = version_name + " " + version_number + "\n"
        msg += "Contributors: Peter Isaac, Jamie Cleverly, Cacilia Ewenz, Ian McHugh"
        # pfp_gui.myMessageBox(msg)
        return

    def help_wiki(self):
        """ Opens the default browser and goes to the PyFluxPro Wiki."""
        browser = webbrowser.get()
        browser.open_new("https://github.com/OzFlux/PyFluxPro/wiki")
        return

    def open_control_file(self):
        logger = logging.getLogger(f"{version.__name__}_log")
        # check to see if the processing level is defined in the control file
        if "level" not in self.file:
            # if not, then sniff the control file to see what it is
            self.file["level"] = self.get_cf_level()
            # and save the control file
            self.file.write()
        # create a QtTreeView to edit the control file
        # if self.file["level"] in ["L1"]:
        #     # update control file to new syntax
        #     if not pfp_compliance.l1_update_controlfile(self.file):
        #         return
        #     # put the GUI for editing the L1 control file in a new tab
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_L1(
        #         self)
        # elif self.file["level"] in ["L2"]:
        #     if not pfp_compliance.l2_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_L2(
        #         self)
        # elif self.file["level"] in ["L3"]:
        #     if not pfp_compliance.l3_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_L3(
        #         self)
        # elif self.file["level"] in ["concatenate"]:
        #     if not pfp_compliance.concatenate_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_concatenate(
        #         self)
        # elif self.file["level"] in ["climatology"]:
        #     if not pfp_compliance.climatology_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_climatology(
        #         self)
        # elif self.file["level"] in ["cpd_barr"]:
        #     if not pfp_compliance.cpd_barr_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_cpd_barr(
        #         self)
        # elif self.file["level"] in ["cpd_mchugh"]:
        #     if not pfp_compliance.cpd_mchugh_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_cpd_mchugh(
        #         self)
        # elif self.file["level"] in ["cpd_mcnew"]:
        #     if not pfp_compliance.cpd_mcnew_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_cpd_mcnew(
        #         self)
        # elif self.file["level"] in ["mpt"]:
        #     if not pfp_compliance.mpt_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_mpt(
        #         self)
        # elif self.file["level"] in ["L4"]:
        #     if not pfp_compliance.l4_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_L4(
        #         self)
        # elif self.file["level"] in ["L5"]:
        #     if not pfp_compliance.l5_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_L5(
        #         self)
        # elif self.file["level"] in ["L6"]:
        #     if not pfp_compliance.l6_update_controlfile(self.file):
        #         return
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_L6(
        #         self)
        # elif self.file["level"] in ["nc2csv_biomet"]:
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_nc2csv_biomet(
        #         self)
        # elif self.file["level"] in ["nc2csv_ecostress"]:
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_nc2csv_ecostress(
        #         self)
        # elif self.file["level"] in ["nc2csv_fluxnet"]:
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_nc2csv_fluxnet(
        #         self)
        # elif self.file["level"] in ["nc2csv_reddyproc"]:
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_nc2csv_reddyproc(
        #         self)
        # elif self.file["level"] in ["batch"]:
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_batch(
        #         self)
        # elif self.file["level"] in ["windrose"]:
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_windrose(
        #         self)
        # elif self.file["level"] in ["PP"]:
        #     self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.edit_cfg_PP(
        #         self)
        # else:
        #     logger.error(" Unrecognised control file type: " +
        #                  self.file["level"])
        #     return

        self.tabs.tab_dict[self.tabs.tab_index_all] = gui.edit_cfg_setup(
            self)
        # add a tab for the control file
        tab_title = os.path.basename(str(self.file_uri))
        self.tabs.addTab(
            self.tabs.tab_dict[self.tabs.tab_index_all], tab_title)
        self.tabs.setCurrentIndex(self.tabs.tab_index_all)
        self.tabs.tab_index_all = self.tabs.tab_index_all + 1
        return

    def open_netcdf_file(self):
        file_uri = self.file.filepath()
        # close the netCDF file
        self.file.close()
        # read the netCDF file to a data structure
        self.ds = pfp_io.NetCDFRead(file_uri, checktimestep=False)
        if self.ds.info["returncodes"]["value"] != 0:
            return
        # display the netcdf file in the GUI
        self.tabs.tab_dict[self.tabs.tab_index_all] = pfp_gui.file_explore(
            self)
        # return if something went wrong
        if self.tabs.tab_dict[self.tabs.tab_index_all].ds.info["returncodes"]["value"] != 0:
            return
        # add a tab for the netCDF file contents
        tab_title = os.path.basename(self.ds.info["filepath"])
        self.tabs.addTab(
            self.tabs.tab_dict[self.tabs.tab_index_all], tab_title)
        self.tabs.setCurrentIndex(self.tabs.tab_index_all)
        self.tabs.tab_index_all = self.tabs.tab_index_all + 1
        return

    def get_cf_level(self):
        """ Sniff the control file to find out it's type."""
        logger = logging.getLogger(f"{version.__name__}_log")
        self.file["level"] = ""
        # check for L1
        if self.check_cfg_L1():
            logger.info(" L1 control file detected")
            self.file["level"] = "L1"
        # check for L2
        elif self.check_cfg_L2():
            logger.info(" L2 control file detected")
            self.file["level"] = "L2"
        # check for L3
        elif self.check_cfg_L3():
            logger.info(" L3 control file detected")
            self.file["level"] = "L3"
        # check for concatenate
        elif self.check_cfg_concatenate():
            logger.info(" Concatenate control file detected")
            self.file["level"] = "concatenate"
        # check for L4
        elif self.check_cfg_L4():
            logger.info(" L4 control file detected")
            self.file["level"] = "L4"
        # check for L5
        elif self.check_cfg_L5():
            logger.info(" L5 control file detected")
            self.file["level"] = "L5"
        # check for L6
        elif self.check_cfg_L6():
            logger.info(" L6 control file detected")
            self.file["level"] = "L6"
        else:
            logger.info(" Unable to detect level, enter manually ...")
            text, ok = QInputDialog.getText(
                self, 'Processing level', 'Enter the processing level:')
            if ok:
                self.file["level"] = text
        return self.file["level"]

    def check_cfg_L1(self):
        """ Return true if a control file is an L1 file."""
        result = False
        try:
            cfg_sections = list(self.file.keys())
            # remove the common sections
            common_sections = ["level", "controlfile_name", "Files", "Global", "Output",
                               "Plots", "General", "Options", "Soil", "Massman", "GUI"]
            for section in list(self.file.keys()):
                if section in common_sections:
                    cfg_sections.remove(section)
            # now loop over the remaining sections
            for section in cfg_sections:
                subsections = list(self.file[section].keys())
                for subsection in subsections:
                    if "Attr" in list(self.file[section][subsection].keys()):
                        result = True
                        break
        except:
            result = False
        return result

    def check_cfg_L2(self):
        """ Return true if a control file is an L2 file."""
        result = False
        try:
            got_sections = False
            cfg_sections = list(self.file.keys())
            if (("Files" in cfg_sections) and
                    ("Variables" in cfg_sections)):
                got_sections = True
            # loop over [Variables] sections
            got_qc = False
            qc_list = ["RangeCheck", "DiurnalCheck", "ExcludeDates", "DependencyCheck", "UpperCheck",
                       "LowerCheck", "ExcludeHours", "Linear", "CorrectWindDirection"]
            for section in ["Variables"]:
                subsections = list(self.file[section].keys())
                for subsection in subsections:
                    for qc in qc_list:
                        if qc in list(self.file[section][subsection].keys()):
                            got_qc = True
                            break
            # final check
            if got_sections and got_qc and not self.check_cfg_L3() and not self.check_cfg_L4():
                result = True
        except:
            result = False
        return result

    def check_cfg_L3(self):
        """ Return true if a control file is an L3 file."""
        result = False
        try:
            cfg_sections = list(self.file.keys())
            if ((("General" in cfg_sections) or
                ("Soil" in cfg_sections) or
                ("Massman" in cfg_sections)) and
                    ("Options" in cfg_sections)):
                result = True
        except:
            result = False
        return result

    def check_cfg_concatenate(self):
        """ Return true if control file is concatenation."""
        result = False
        try:
            cfg_sections = list(self.file.keys())
            if "Files" in cfg_sections:
                if (("Out" in list(self.file["Files"].keys())) and
                        ("In" in list(self.file["Files"].keys()))):
                    result = True
        except:
            result = False
        return result

    def check_cfg_L4(self):
        """ Return true if control file is L4."""
        result = False
        try:
            cfg_sections = list(self.file.keys())
            for section in cfg_sections:
                if section in ["Variables", "Drivers", "Fluxes"]:
                    subsections = list(self.file[section].keys())
                    for subsection in subsections:
                        if (("GapFillFromAlternate" in list(self.file[section][subsection].keys())) or
                                ("GapFillFromClimatology" in list(self.file[section][subsection].keys()))):
                            result = True
                            break
        except:
            result = False
        return result

    def check_cfg_L5(self):
        """ Return true if control file is L5."""
        result = False
        try:
            cfg_sections = list(self.file.keys())
            for section in cfg_sections:
                if section in ["Variables", "Drivers", "Fluxes"]:
                    subsections = list(self.file[section].keys())
                    for subsection in subsections:
                        if (("GapFillUsingSOLO" in list(self.file[section][subsection].keys())) or
                                ("GapFillUsingMDS" in list(self.file[section][subsection].keys()))):
                            result = True
                            break
        except:
            result = False
        return result

    def check_cfg_L6(self):
        """ Return true if control file is L6."""
        result = False
        try:
            cfg_sections = list(self.file.keys())
            if ("EcosystemRespiration" in cfg_sections or
                "NetEcosystemExchange" in cfg_sections or
                    "GrossPrimaryProductivity" in cfg_sections):
                result = True
        except:
            result = False
        return result

    def direct_run(self):
        """ Placeholder until full implementation done."""
        logger = logging.getLogger(f"{version.__name__}_log")
        msg = " Open control file and use 'Run/Current ...'"
        logger.warning(msg)
        return

    def file_save(self):
        """ Save the current file."""
        logger = logging.getLogger(f"{version.__name__}_log")
        # get the current tab index
        tab_index_current = self.tabs.tab_index_current
        # trap user attempts to save the Log window :=) thanks, Craig Macfarlane!
        if tab_index_current == 0:
            msg = "Log files are automatically saved" + "\n"
            msg += "in the PyFluxPro" + os.sep + "logfiles folder"
            pfp_gui.MsgBox_Continue(msg)
            return
        content = self.tabs.tab_dict[tab_index_current].get_data_from_model()
        if isinstance(content, ConfigObj):
            # we are saving a control file
            self.save_control_file()
        elif isinstance(content, pfp_io.DataStructure):
            # we are saving a data structure
            self.save_netcdf_file()
        else:
            # unrecognised content type
            msg = "Object must be either a control file or a data structure"
            logger.error(msg)
        return

    def save_control_file(self):
        """ Save the current tab as a control file."""
        logger = logging.getLogger(f"{version.__name__}_log")
        # get the current tab index
        tab_index_current = self.tabs.tab_index_current
        # get the updated control file data
        cfg = self.tabs.tab_dict[tab_index_current].get_data_from_model()
        # check to make sure we are not overwriting the template version
        if "template" in cfg.filename:
            msg = " You are trying to write to the template folder.\n"
            msg = msg + "Please save this control file to a different location."
            pfp_gui.myMessageBox(msg)
            # put up a "Save as ..." dialog
            cfg_filename = QFileDialog.getSaveFileName(self, "Save as ...")[0]
            # return without doing anything if cancel used
            if len(str(cfg_filename)) == 0:
                return
            # set the control file name
            cfg.filename = str(cfg_filename)
        # write the control file
        msg = " Saving " + cfg.filename
        logger.info(msg)
        cfg.write()
        # remove the asterisk in the tab text
        tab_text = str(self.tabs.tabText(tab_index_current))
        self.tabs.setTabText(self.tabs.tab_index_current,
                             tab_text.replace("*", ""))
        return

    def save_netcdf_file(self):
        """Save the current tab as a netCDF file."""
        # get the current tab index
        tab_index_current = self.tabs.tab_index_current
        # get the updated control file data
        ds = self.tabs.tab_dict[tab_index_current].get_data_from_model()
        # write the data structure to file
        pfp_io.NetCDFWrite(ds.info["filepath"], ds)
        # remove the asterisk in the tab text
        tab_text = str(self.tabs.tabText(tab_index_current))
        self.tabs.setTabText(self.tabs.tab_index_current,
                             tab_text.replace("*", ""))
        return

    def file_save_as(self):
        """Save a file with a different name."""
        logger = logging.getLogger(f"{version.__name__}_log")
        # get the current tab index
        tab_index_current = self.tabs.tab_index_current
        # trap user attempts to save the Log window :=) thanks, Craig Macfarlane!
        if tab_index_current == 0:
            msg = "Log files are automatically saved" + "\n"
            msg += "in the PyFluxPro" + os.sep + "logfiles folder"
            pfp_gui.MsgBox_Continue(msg)
            return
        content = self.tabs.tab_dict[tab_index_current].get_data_from_model()
        if isinstance(content, ConfigObj):
            # we are saving a control file
            file_uri = content.filename
            file_uri = QFileDialog.getSaveFileName(
                self, "Save as ...", file_uri)[0]
            if len(str(file_uri)) == 0:
                return
            content.filename = file_uri
            self.save_as_control_file(content)
        elif isinstance(content, pfp_io.DataStructure):
            # we are opening a netCDF file
            file_uri = content.info["filepath"]
            file_uri = QFileDialog.getSaveFileName(
                self, "Save as ...", file_uri)[0]
            if len(str(file_uri)) == 0:
                return
            content.info["filepath"] = file_uri
            self.save_as_netcdf_file(content)
        else:
            # unrecognised file type
            msg = "File must be either a control file or a netCDF file"
            logger.error(msg)
        return

    def save_as_control_file(self, cfg):
        """ Save the current tab with a different name."""
        tab_index_current = self.tabs.tab_index_current
        logger = logging.getLogger(f"{version.__name__}_log")
        # write the control file
        logger.info(" Saving " + cfg.filename)
        cfg.write()
        # update the tab text
        tab_title = os.path.basename(str(cfg.filename))
        self.tabs.setTabText(tab_index_current, tab_title)
        return

    def save_as_netcdf_file(self, ds):
        """ Save the current tab with a different name."""
        # get the current tab index
        tab_index_current = self.tabs.tab_index_current
        # write the data structure to file
        pfp_io.NetCDFWrite(ds.info["filepath"], ds)
        # update the tab text
        tab_title = os.path.basename(str(ds.info["filepath"]))
        self.tabs.setTabText(tab_index_current, tab_title)
        return

    def edit_preferences(self):
        logger.debug("Edit/Preferences goes here")
        pass

    def tabSelected(self, arg=None):
        self.tabs.tab_index_current = arg

    def run_clear_log_window(self):
        """ Clear the Log window"""
        # the Log window is always tab index 0
        tb = self.tabs.widget(0)
        tb.clear()

    def run_current(self):
        self.stop_flag = False
        # save the current tab index
        logger = logging.getLogger(f"{version.__name__}_log")
        tab_index_current = self.tabs.tab_index_current
        if tab_index_current == 0:
            msg = " No control file selected ..."
            logger.warning(msg)
            return
        # disable the Run/Current menu option
        self.actionRunCurrent.setDisabled(True)
        # get the updated control file data
        cfg = self.tabs.tab_dict[tab_index_current].get_data_from_model()
        # set the focus back to the log tab
        self.tabs.setCurrentIndex(0)
        # call the appropriate processing routine depending on the level
        self.tabs.tab_index_running = tab_index_current
        if cfg["level"] == "batch":
            # check the L1 control file to see if it is OK to run
            if not pfp_compliance.check_batch_controlfile(cfg):
                return
            # add stop to run menu
            self.menuRun.addAction(self.actionStopCurrent)
            self.actionStopCurrent.triggered.connect(self.stop_current)
            # get a worker thread
            # worker = pfp_threading.Worker(pfp_top_level.do_run_batch, self)
            # start the worker
            self.threadpool.start(worker)
            # no threading
            # pfp_top_level.do_run_batch(self)
        # elif cfg["level"] == "L1":
        #     # check the L1 control file to see if it is OK to run
        #     if pfp_compliance.check_l1_controlfile(cfg):
        #         pfp_top_level.do_run_l1(cfg)
        #     self.actionRunCurrent.setDisabled(False)
        # elif cfg["level"] == "L2":
        #     # check the L2 control file to see if it is OK to run
        #     if pfp_compliance.check_l2_controlfile(cfg):
        #         pfp_top_level.do_run_l2(cfg)
        #     self.actionRunCurrent.setDisabled(False)
        # elif cfg["level"] == "L3":
        #     pfp_top_level.do_run_l3(cfg)
        #     self.actionRunCurrent.setDisabled(False)
        # elif cfg["level"] == "concatenate":
        #     pfp_top_level.do_file_concatenate(cfg)
            self.actionRunCurrent.setDisabled(False)
        elif cfg["level"] == "climatology":
            self.utilities_climatology_custom()
        elif cfg["level"] == "cpd_barr":
            self.utilities_ustar_cpd_barr_custom()
        elif cfg["level"] == "cpd_mchugh":
            self.utilities_ustar_cpd_mchugh_custom()
        elif cfg["level"] == "cpd_mcnew":
            self.utilities_ustar_cpd_mcnew_custom()
        elif cfg["level"] == "mpt":
            self.utilities_ustar_mpt_custom()
        # elif cfg["level"] == "L4":
        #     pfp_top_level.do_run_l4(self)
        #     self.actionRunCurrent.setDisabled(False)
        # elif cfg["level"] == "L5":
        #     if pfp_compliance.check_l5_controlfile(cfg):
        #         pfp_top_level.do_run_l5(self)
        #     self.actionRunCurrent.setDisabled(False)
        # elif cfg["level"] == "L6":
        #     if pfp_compliance.check_l6_controlfile(cfg):
        #         pfp_top_level.do_run_l6(self)
        #     self.actionRunCurrent.setDisabled(False)
        # elif cfg["level"] == "nc2csv_biomet":
        #     pfp_top_level.do_file_convert_nc2biomet(self, mode="custom")
        #     self.actionRunCurrent.setDisabled(False)
        # elif cfg["level"] == "nc2csv_ecostress":
        #     pfp_top_level.do_file_convert_nc2ecostress(self)
        #     self.actionRunCurrent.setDisabled(False)
        # elif cfg["level"] == "nc2csv_fluxnet":
        #     pfp_top_level.do_file_convert_nc2fluxnet(self)
        #     self.actionRunCurrent.setDisabled(False)
        # elif cfg["level"] == "nc2csv_reddyproc":
        #     pfp_top_level.do_file_convert_nc2reddyproc(self, mode="custom")
        #     self.actionRunCurrent.setDisabled(False)
        # elif cfg["level"] == "windrose":
        #     if pfp_compliance.check_windrose_controlfile(cfg):
        #         pfp_top_level.do_plot_windrose_custom(self)
        #     self.actionRunCurrent.setDisabled(False)
        elif cfg["level"] == "PP":
            # add stop to run menu
            self.menuRun.addAction(self.actionStopCurrent)
            self.actionStopCurrent.triggered.connect(self.stop_current)
            # get a worker thread

            def job(self_):
                opf_processing.do_run_preprocess(self_)
                self_.actionRunCurrent.setDisabled(False)
                self_.menuRun.removeAction(self.actionStopCurrent)
            worker = pfp_threading.Worker(job, self)
            # start the worker
            self.threadpool.start(worker)
        else:
            logger.error("Level not implemented yet ...")
        return

    def stop_current(self):
        # put up a message box, continue or quit
        msg = "Do you want to quit processing?"
        result = pfp_gui.MsgBox_ContinueOrQuit(msg, title="Quit processing?")
        if result.clickedButton().text() == "Quit":
            self.stop_flag = True
            msg = "Processing will stop when this level is finished"
            result = pfp_gui.myMessageBox(msg)
        return

    def closeTab(self, currentIndex):
        """ Close the selected tab."""
        # the tab close button ("x") shows on MacOS even though it is disabled
        # here we trap user attempts to close the log window
        if (currentIndex == 0):
            return
        # check to see if the tab contents have been saved
        tab_text = str(self.tabs.tabText(currentIndex))
        if "*" in tab_text:
            msg = "Save control file?"
            reply = QMessageBox.question(self, 'Message', msg,
                                               QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
            if reply == QMessageBox.Yes:
                self.file_save()
            if reply == QMessageBox.Cancel:
                return
        # get the current tab from its index
        currentQWidget = self.tabs.widget(currentIndex)
        # delete the tab
        currentQWidget.deleteLater()
        self.tabs.removeTab(currentIndex)
        # remove the corresponding entry in tab_dict
        self.tabs.tab_dict.pop(currentIndex)
        # and renumber the keys
        for n in list(self.tabs.tab_dict.keys()):
            if n > currentIndex:
                self.tabs.tab_dict[n-1] = self.tabs.tab_dict.pop(n)
        # decrement the tab index
        self.tabs.tab_index_all = self.tabs.tab_index_all - 1
        return

    def update_tab_text(self):
        """ Add an asterisk to the tab title text to indicate tab contents have changed."""
        # add an asterisk to the tab text to indicate the tab contents have changed
        tab_text = str(self.tabs.tabText(self.tabs.tab_index_current))
        if "*" not in tab_text:
            self.tabs.setTabText(self.tabs.tab_index_current, tab_text+"*")
        return


if (__name__ == '__main__'):
    # get the application name and version
    pfp_version = version_name + " " + version_number
    app = QApplication(["PyFluxPro"])
    # get a text editor for the log window
    textBox = QTextEdit()
    textBox.setReadOnly(True)
    logger.consoleHandler.sigLog.connect(textBox.append)
    # instance the main GUI
    ui = pfp_main_ui(pfp_version, textBox)
    ui.show()
    # pfp_compliance.check_executables()
    app.exec_()
    del ui



"""This module implements the double, triple rotation and planar fit corrections for wind components
as described by Wilczak et al. (2001). The corrections are applied to the u, v, and w components of wind data.
It provides functions for both double and triple rotation corrections, allowing for the correction of wind data
to account for the tilt of the anemometer and the alignment of the mean wind vector.
"""

# built-in modules
import re
import os
import warnings
import logging
from functools import reduce

# 3rd party modules
import numpy as np
import pandas as pd
import scipy as sp

# project modules

logger = logging.getLogger('ep.corrections.axis_rotation.wilczak_et_al_2001')

def double_rotation(u, v, w, verbosity=0):
    """
    Perform a double rotation correction on the input wind components u, v, w.
    This function applies the first rotation to align the mean wind vector with the x-axis,
    and the second rotation to align the mean vertical wind component with the z-axis.
    Parameters:
        u (array-like): The u component of the wind.
        v (array-like): The v component of the wind.
        w (array-like): The w component of the wind.
        verbosity (int): Level of verbosity for debug output.
    Returns:
        tuple: Corrected u, v, w components and the angles of the first (theta) and second (phi) rotations.
    """
    # Ensure inputs are numpy arrays
    u = np.asarray(u)
    v = np.asarray(v)
    w = np.asarray(w)

    #first rotation
    theta = np.arctan(np.nanmean(v)/np.nanmean(u))
    u1 = u * np.cos(theta) + v * np.sin(theta)
    v1 = -u * np.sin(theta) + v * np.cos(theta)
    w1 = w

    #second rotation
    phi = np.arctan(np.nanmean(w1)/np.nanmean(u1))
    u2 = u1 * np.cos(phi) + w1 * np.sin(phi)
    v2 = v1
    w2 = -u1 * np.sin(phi) + w1 * np.cos(phi)
    
    if verbosity > 0:
        print(f"Double rotation angles: theta={theta}, phi={phi}")
    
    # Return the corrected components and angles as a named tuple-like object
    return type('var_', (object,), 
                {"u": u2, 
                 "v": v2, 
                 "w": w2, 
                 "theta": theta,
                 "phi": phi,
                 "meta": {}})


def triple_rotation(u, v, w, verbosity=0):
    """
    Perform a triple rotation correction on the input wind components u, v, w.
    This function applies the double rotation method followed by a third rotation
    to correct for the tilt of the anemometer.
    Parameters:
        u (array-like): The u component of the wind.
        v (array-like): The v component of the wind.
        w (array-like): The w component of the wind.
        verbosity (int): Level of verbosity for debug output.
    Returns:
        tuple: Corrected u, v, w components and the angle of the third rotation (psi).
    """
    #first and second rotations
    u2, v2, w2, theta, phi = double_rotation(u, v, w, verbosity)
    
    #third rotation
    psi = np.arctan((2 * np.nanmean(v2 * w2)) /
                        (np.nanmean(v2**2) - np.nanmean(w2**2)))
    u3 = u2
    v3 = v2 * np.cos(psi) + w2 * np.sin(psi)
    w3 = -v2 * np.sin(psi) + w2 * np.cos(psi)
    
    if verbosity > 0:
        print(f"First rotation angles: theta={theta}, phi={phi}, psi={psi}")
    if verbosity > 0:
        print(f"Mean w after triple rotation: {np.nanmean(w3)}")

    # Return the corrected components and angles as a named tuple-like object
    return type('var_', (object,), 
                {"u": u3, 
                 "v": v3, 
                 "w": w3, 
                 "theta": theta, 
                 "phi": phi, 
                 "psi": psi,
                 "meta": {}})


def planarfit(u, v, w, verbosity=0):
    """Perform a planar fit correction on the input wind components u, v, w.
    This function calculates the mean wind vector and applies a rotation to align it with the z-axis.
    Parameters:
        u (array-like): The u component of the wind.
        v (array-like): The v component of the wind.
        w (array-like): The w component of the wind.
        verbosity (int): Level of verbosity for debug output.
    Returns:
        tuple: Corrected u, v, w components after planar fit correction.
    """

    # Ensure inputs are numpy arrays
    u = np.asarray(u)
    v = np.asarray(v)
    w = np.asarray(w)

    meanU = np.nanmean(u)
    meanV = np.nanmean(v)
    meanW = np.nanmean(w)

    def findB(meanU, meanV, meanW):
        su = np.nansum(meanU)
        sv = np.nansum(meanV)
        sw = np.nansum(meanW)

        suv = meanU * meanV
        suw = meanU * meanW
        svw = meanV * meanW
        su2 = meanU * meanU
        sv2 = meanV * meanV

        H = np.matrix([[1, su, sv], [su, su2, suv], [sv, suv, sv2]])
        g = np.matrix([sw, suw, svw]).T
        x = sp.linalg.solve(H, g)

        b0 = x[0][0]
        b1 = x[1][0]
        b2 = x[2][0]
        return b0, b1, b2

    b0, b1, b2 = findB(meanU, meanV, meanW)

    Deno = np.sqrt(1 + b1 ** 2 + b2 ** 2)
    p31 = -b1 / Deno
    p32 = -b2 / Deno
    p33 = 1 / Deno

    cosγ = p33 / np.sqrt(p32**2+p33**2)
    sinγ = -p32 / np.sqrt(p32**2 + p33**2)
    cosβ = np.sqrt(p32**2 + p33**2)
    sinβ = p31

    R2 = np.matrix([[1, 0, 0],
                    [0, cosγ, -sinγ],
                    [0, sinγ, cosγ]])
    R3 = np.matrix([[cosβ, 0, sinβ],
                    [0, 1, 0],
                    [-sinβ, 0, cosβ]])

    A0 = R3.T * R2.T * [[meanU], [meanV], [meanW]]

    α = np.arctan2(A0[1].tolist()[0][0],
                   A0[0].tolist()[0][0])

    R1 = np.matrix([[np.cos(α), -np.sin(α), 0],
                    [np.sin(α), np.cos(α), 0],
                    [0, 0, 1]])

    A1 = R1.T * ((R3.T * R2.T) * np.matrix([u, v, w - b0]))

    U1 = np.array(A1[0])[0]
    V1 = np.array(A1[1])[0]
    W1 = np.array(A1[2])[0]

    if verbosity > 0:
        print(f"Planar fit angles: α={α}, β={np.arctan2(sinβ, cosβ)}, γ={np.arctan2(sinγ, cosγ)}")
        print(f"Mean w after planar fit: {np.nanmean(W1)}")

    if type(u) == pd.Series:
        U1 = pd.Series(U1)
    if type(v) == pd.Series:
        V1 = pd.Series(V1)
    if type(w) == pd.Series:
        W1 = pd.Series(W1)

    # Return the corrected components as a named tuple-like object
    return type('var_', (object,), 
                {"u": U1, 
                 "v": V1, 
                 "w": W1,
                 "meta": {}})




# built-in modules
from configobj import ConfigObj
import re
import os
import warnings
import logging
from functools import reduce

# 3rd party modules
import numpy as np
import pandas as pd

# project modules

logger = logging.getLogger('wvlt.pyfluxpro_compatibility')

def copy_values_from(source, target, overwrite=True):
    result = {}
    for key, val in source.items():
        if isinstance(val, dict):
            # Recurse into nested dict
            result[key] = copy_values_from(val, target.get(key, {}))
        else:
            if overwrite:
                # Copy value from target if exists, else keep original
                result[key] = target.get(key, val)
            else:
                result[key] = val or target.get(key, val)
    return result

def find_info_to_fill(path):
    return set(re.findall('<.*?>', str(ConfigObj(path))))

def L1_cfg(config,
           dst='eddypy/compatibility/PyFluxPro/L1.txt',
           template='eddypy/compatibility/PyFluxPro/templates/L1/L1_eddypro_csv.txt',
           **options):
    with open(dst, 'w+') as wf:
        with open(template, 'r') as rf:
            for line in rf.readlines():
                for k, v in options.items():
                    line = line.replace(f'<{k}>', v)
                wf.write(line)

    setup = ConfigObj(dst)
    setup['Files'] = config['Files']
    setup.write()
    return


def L2_cfg(config, 
           dst='eddypy/compatibility/PyFluxPro/L2.txt', 
           template='eddypy/compatibility/PyFluxPro/templates/L2/L2_eddypro.txt'):
    with open(dst, 'w+') as wf:
        with open(template, 'r') as rf:
            for line in rf.readlines():
                # for k, v in config['Options'].items():
                #     line = line.replace(f'<{k}>', v)
                wf.write(line)

    setup = ConfigObj(dst)
    setup['Files'] = config['Files']
    setup['Options'] = config['Options']
    setup.write()
    return

    
def L3_cfg(config, 
           dst='eddypy/compatibility/PyFluxPro/L3.txt', 
           template='eddypy/compatibility/PyFluxPro/templates/L3/L3_eddypro.txt'):
    with open(dst, 'w+') as wf:
        with open(template, 'r') as rf:
            for line in rf.readlines():
                for k, v in config['Options'].items():
                    line = line.replace(f'<{k}>', v)
                wf.write(line)

    setup = ConfigObj(dst)
    setup['Files'] = config['Files']
    setup['Options'] = config['Options']
    setup.write()
    return


def L4_cfg(config,
           dst='eddypy/compatibility/PyFluxPro/L4.txt',
           template='eddypy/compatibility/PyFluxPro/templates/L4/L4_fluxnet.txt'):
    with open(dst, 'w+') as wf:
        with open(template, 'r') as rf:
            for line in rf.readlines():
                for k, v in config['Options'].items():
                    line = line.replace(f'<{k}>', v)
                wf.write(line)

    setup = ConfigObj(dst)
    setup['Files'] = config['Files']
    setup['Options'] = config['Options']
    setup.write()
    return


def L5_cfg(config,
           dst='eddypy/compatibility/PyFluxPro/L5.txt',
           template='eddypy/compatibility/PyFluxPro/templates/L5/L5_MDS.txt'):
    with open(dst, 'w+') as wf:
        with open(template, 'r') as rf:
            for line in rf.readlines():
                for k, v in config['Options'].items():
                    line = line.replace(f'<{k}>', v)
                wf.write(line)

    setup = ConfigObj(dst)
    setup['Files'] = config['Files']
    setup['Options'] = config['Options']
    setup.write()
    return
